#if 0
g++ -std=gnu++11 ropgenerate.cpp udis86-1.7.2/libudis86/.libs/libudis86.a -oropgenerate

./ropgenerate zsneswu.exe \
  | LC_ALL=C sort -k2 | uniq -f2 | cut -d' ' -f1,3- \
  > zsneswu-rop.txt
./ropgenerate zsnesw_8M.exe \
  | LC_ALL=C sort -k2 | uniq -f2 | cut -d' ' -f1,3- \
  > zsnesw_8M-rop.txt
exit
#endif

#include "udis86-1.7.2/udis86.h"
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

void printhex(const uint8_t* data, size_t len)
{
	for (size_t i=0;i<len;i++) printf("%.2X%c",data[i],(i==len-1)?'\n':' ');
}

void disassemble_section(const uint8_t* data, size_t len, size_t offset, uint8_t bits)
{
	ud_t ud_obj;
	
	ud_init(&ud_obj);
	ud_set_mode(&ud_obj, 32);
	ud_set_syntax(&ud_obj, UD_SYN_INTEL);
	
	for (int j=1;j<16;j++)
	{
		for (int i=j;i<len;i++)
		{
			if (data[i]==0xC3)//ret
			{
				uint32_t start = offset+i-j+1;
				//ud_set_pc(&ud_obj, offset);
				ud_set_pc(&ud_obj, 0);//doing this makes 'uniq' kill 90% of the JMPs
				ud_set_input_buffer(&ud_obj, data+i-j+1, j);
				
				char text[4096];
				char* textend=text;
				
				int num_ops=0;
				while (ud_disassemble(&ud_obj)) {
					num_ops++;
					
					const char * op = ud_insn_asm(&ud_obj);
					textend+=sprintf(textend, "%s; ", op);
					if (!strcmp(op, "invalid")) break;
					
					if (!strcmp(op, "ret"))
					{
						if (ud_obj.inp_buf_size==ud_obj.inp_buf_index)
						{
							textend[-2]='\0';
							printf("%.8X: %.2d %s\n", start, num_ops, text);
						}
						break;
					}
				}
			}
		}
	}
}

void disassemble_windows(uint8_t* binary, size_t len)
{
	typedef uint8_t UCHAR;
	typedef uint16_t USHORT;
	typedef uint32_t ULONG;
	typedef int8_t CHAR;
	typedef int16_t SHORT;
	typedef int32_t LONG;
	
	typedef struct _IMAGE_DOS_HEADER {  // DOS .EXE header
	    USHORT e_magic;         // Magic number
	    USHORT e_cblp;          // Bytes on last page of file
	    USHORT e_cp;            // Pages in file
	    USHORT e_crlc;          // Relocations
	    USHORT e_cparhdr;       // Size of header in paragraphs
	    USHORT e_minalloc;      // Minimum extra paragraphs needed
	    USHORT e_maxalloc;      // Maximum extra paragraphs needed
	    USHORT e_ss;            // Initial (relative) SS value
	    USHORT e_sp;            // Initial SP value
	    USHORT e_csum;          // Checksum
	    USHORT e_ip;            // Initial IP value
	    USHORT e_cs;            // Initial (relative) CS value
	    USHORT e_lfarlc;        // File address of relocation table
	    USHORT e_ovno;          // Overlay number
	    USHORT e_res[4];        // Reserved words
	    USHORT e_oemid;         // OEM identifier (for e_oeminfo)
	    USHORT e_oeminfo;       // OEM information; e_oemid specific
	    USHORT e_res2[10];      // Reserved words
	    LONG   e_lfanew;        // File address of new exe header
	  } IMAGE_DOS_HEADER, *PIMAGE_DOS_HEADER;
	
	typedef struct _IMAGE_FILE_HEADER {
	    USHORT  Machine;
	    USHORT  NumberOfSections;
	    ULONG   TimeDateStamp;
	    ULONG   PointerToSymbolTable;
	    ULONG   NumberOfSymbols;
	    USHORT  SizeOfOptionalHeader;
	    USHORT  Characteristics;
	} IMAGE_FILE_HEADER, *PIMAGE_FILE_HEADER;
	
	typedef struct _IMAGE_DATA_DIRECTORY {
	    ULONG   VirtualAddress;
	    ULONG   Size;
	} IMAGE_DATA_DIRECTORY, *PIMAGE_DATA_DIRECTORY;
	#define IMAGE_NUMBEROF_DIRECTORY_ENTRIES 16
	
	typedef struct _IMAGE_OPTIONAL_HEADER {
	    //
	    // Standard fields.
	    //
	    USHORT  Magic;
	    UCHAR   MajorLinkerVersion;
	    UCHAR   MinorLinkerVersion;
	    ULONG   SizeOfCode;
	    ULONG   SizeOfInitializedData;
	    ULONG   SizeOfUninitializedData;
	    ULONG   AddressOfEntryPoint;
	    ULONG   BaseOfCode;
	    ULONG   BaseOfData;
	    //
	    // NT additional fields.
	    //
	    ULONG   ImageBase;
	    ULONG   SectionAlignment;
	    ULONG   FileAlignment;
	    USHORT  MajorOperatingSystemVersion;
	    USHORT  MinorOperatingSystemVersion;
	    USHORT  MajorImageVersion;
	    USHORT  MinorImageVersion;
	    USHORT  MajorSubsystemVersion;
	    USHORT  MinorSubsystemVersion;
	    ULONG   Reserved1;
	    ULONG   SizeOfImage;
	    ULONG   SizeOfHeaders;
	    ULONG   CheckSum;
	    USHORT  Subsystem;
	    USHORT  DllCharacteristics;
	    ULONG   SizeOfStackReserve;
	    ULONG   SizeOfStackCommit;
	    ULONG   SizeOfHeapReserve;
	    ULONG   SizeOfHeapCommit;
	    ULONG   LoaderFlags;
	    ULONG   NumberOfRvaAndSizes;
	    IMAGE_DATA_DIRECTORY DataDirectory[IMAGE_NUMBEROF_DIRECTORY_ENTRIES];
	} IMAGE_OPTIONAL_HEADER, *PIMAGE_OPTIONAL_HEADER;
	
	#define IMAGE_SIZEOF_SHORT_NAME              8
	
	#define IMAGE_SCN_MEM_EXECUTE 0x20000000
	#define IMAGE_SCN_MEM_READ 0x40000000
	#define IMAGE_SCN_MEM_WRITE 0x80000000
	
	typedef struct _IMAGE_SECTION_HEADER {
	    UCHAR   Name[IMAGE_SIZEOF_SHORT_NAME];
	    union {
	            ULONG   PhysicalAddress;
	            ULONG   VirtualSize;
	    } Misc;
	    ULONG   VirtualAddress;
	    ULONG   SizeOfRawData;
	    ULONG   PointerToRawData;
	    ULONG   PointerToRelocations;
	    ULONG   PointerToLinenumbers;
	    USHORT  NumberOfRelocations;
	    USHORT  NumberOfLinenumbers;
	    ULONG   Characteristics;
	} IMAGE_SECTION_HEADER, *PIMAGE_SECTION_HEADER;
	
	
	PIMAGE_DOS_HEADER dosheader = (PIMAGE_DOS_HEADER)binary;
	PIMAGE_FILE_HEADER fileheader = (PIMAGE_FILE_HEADER)((uint8_t*)dosheader + dosheader->e_lfanew + 4);
	if (memcmp((uint8_t*)fileheader-4, "PE\0\0", 4)) return;
	PIMAGE_OPTIONAL_HEADER optheader = (PIMAGE_OPTIONAL_HEADER)((uint8_t*)fileheader + sizeof(IMAGE_FILE_HEADER));
	
	uint8_t bits;
	if(0);
	else if (optheader->Magic == 0x010B) bits=32;
	else if (optheader->Magic == 0x020B) bits=64;
	else return;
	
	PIMAGE_SECTION_HEADER sections = (PIMAGE_SECTION_HEADER)((uint8_t*)optheader + sizeof(IMAGE_OPTIONAL_HEADER));
	for (uint16_t i=0;i<fileheader->NumberOfSections;i++)
	{
		if (sections[i].Characteristics & IMAGE_SCN_MEM_EXECUTE)
		{
			disassemble_section(binary + sections[i].PointerToRawData, sections[i].SizeOfRawData,
			                    optheader->ImageBase + sections[i].VirtualAddress, bits);
		}
	}
}

int main(int argc, char * argv[])
{
	if (!argv[1])
	{
		puts("usage: ./mass-dis file\nexample: ./mass-dis xkas.exe");
		return 0;
	}
	
	FILE * f=fopen(argv[1], "rb");
	fseek(f, 0, SEEK_END);
	size_t len=ftell(f);
	fseek(f, 0, SEEK_SET);
	uint8_t* data=(uint8_t*)malloc(len);
	fread(data, 1,len, f);
	fclose(f);
	
	if (data[0]=='M' && data[1]=='Z') disassemble_windows(data, len);
	
	//disassemble_section();
	
	return 0;
}
