#if 0
gcc ropgenerate.cpp udis86-1.7.2/libudis86/.libs/libudis86.a -oropgenerate

./ropgenerate zsneswu.exe \
  | LC_ALL=C sort -k2 | uniq -f2 | cut -d' ' -f1,3- \
  > zsneswu-rop.txt
./ropgenerate zsnesw_8M.exe \
  | LC_ALL=C sort -k2 | uniq -f2 | cut -d' ' -f1,3- \
  > zsnesw_8M-rop.txt
exit
#endif

#include "udis86-1.7.2/udis86.h"
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

void printhex(const uint8_t* data, size_t len)
{
	for (size_t i=0;i<len;i++) printf("%.2X%c",data[i],(i==len-1)?'\n':' ');
}

void disassemble_section(const uint8_t* data, size_t len, size_t offset, uint8_t bits)
{
	ud_t ud_obj;
	
	ud_init(&ud_obj);
	ud_set_mode(&ud_obj, bits);
	ud_set_syntax(&ud_obj, UD_SYN_INTEL);
	
	for (int j=1;j<16;j++)
	{
		for (int i=j;i<len;i++)
		{
			if (data[i]==0xC3)//ret
			{
				uint32_t start = offset+i-j+1;
				//ud_set_pc(&ud_obj, offset);
				ud_set_pc(&ud_obj, 0);//doing this makes 'uniq' kill 90% of the JMPs
				ud_set_input_buffer(&ud_obj, data+i-j+1, j);
				
				char text[4096];
				char* textend=text;
				
				int num_ops=0;
				while (ud_disassemble(&ud_obj)) {
					num_ops++;
					
					const char * op = ud_insn_asm(&ud_obj);
					textend+=sprintf(textend, "%s; ", op);
					if (!strcmp(op, "invalid")) break;
					
					if (!strcmp(op, "ret"))
					{
						if (ud_obj.inp_buf_size==ud_obj.inp_buf_index)
						{
							textend[-2]='\0';
							printf("%.8X: %.2d %s\n", start, num_ops, text);
						}
						break;
					}
				}
			}
			
			//if (j==1 && data[i]==0xCD && data[i+1]==0x80)
			//{
			//	printf("%.8X: 0 int 0x80\n", offset+i);
			//}
		}
	}
}



void disassemble_windows(uint8_t* binary, size_t len)
{
	typedef uint8_t UCHAR;
	typedef uint16_t USHORT;
	typedef uint32_t ULONG;
	typedef int8_t CHAR;
	typedef int16_t SHORT;
	typedef int32_t LONG;
	
	typedef struct _IMAGE_DOS_HEADER {  // DOS .EXE header
	    USHORT e_magic;         // Magic number
	    USHORT e_cblp;          // Bytes on last page of file
	    USHORT e_cp;            // Pages in file
	    USHORT e_crlc;          // Relocations
	    USHORT e_cparhdr;       // Size of header in paragraphs
	    USHORT e_minalloc;      // Minimum extra paragraphs needed
	    USHORT e_maxalloc;      // Maximum extra paragraphs needed
	    USHORT e_ss;            // Initial (relative) SS value
	    USHORT e_sp;            // Initial SP value
	    USHORT e_csum;          // Checksum
	    USHORT e_ip;            // Initial IP value
	    USHORT e_cs;            // Initial (relative) CS value
	    USHORT e_lfarlc;        // File address of relocation table
	    USHORT e_ovno;          // Overlay number
	    USHORT e_res[4];        // Reserved words
	    USHORT e_oemid;         // OEM identifier (for e_oeminfo)
	    USHORT e_oeminfo;       // OEM information; e_oemid specific
	    USHORT e_res2[10];      // Reserved words
	    LONG   e_lfanew;        // File address of new exe header
	  } IMAGE_DOS_HEADER, *PIMAGE_DOS_HEADER;
	
	typedef struct _IMAGE_FILE_HEADER {
	    USHORT  Machine;
	    USHORT  NumberOfSections;
	    ULONG   TimeDateStamp;
	    ULONG   PointerToSymbolTable;
	    ULONG   NumberOfSymbols;
	    USHORT  SizeOfOptionalHeader;
	    USHORT  Characteristics;
	} IMAGE_FILE_HEADER, *PIMAGE_FILE_HEADER;
	
	typedef struct _IMAGE_DATA_DIRECTORY {
	    ULONG   VirtualAddress;
	    ULONG   Size;
	} IMAGE_DATA_DIRECTORY, *PIMAGE_DATA_DIRECTORY;
	#define IMAGE_NUMBEROF_DIRECTORY_ENTRIES 16
	
	typedef struct _IMAGE_OPTIONAL_HEADER {
	    //
	    // Standard fields.
	    //
	    USHORT  Magic;
	    UCHAR   MajorLinkerVersion;
	    UCHAR   MinorLinkerVersion;
	    ULONG   SizeOfCode;
	    ULONG   SizeOfInitializedData;
	    ULONG   SizeOfUninitializedData;
	    ULONG   AddressOfEntryPoint;
	    ULONG   BaseOfCode;
	    ULONG   BaseOfData;
	    //
	    // NT additional fields.
	    //
	    ULONG   ImageBase;
	    ULONG   SectionAlignment;
	    ULONG   FileAlignment;
	    USHORT  MajorOperatingSystemVersion;
	    USHORT  MinorOperatingSystemVersion;
	    USHORT  MajorImageVersion;
	    USHORT  MinorImageVersion;
	    USHORT  MajorSubsystemVersion;
	    USHORT  MinorSubsystemVersion;
	    ULONG   Reserved1;
	    ULONG   SizeOfImage;
	    ULONG   SizeOfHeaders;
	    ULONG   CheckSum;
	    USHORT  Subsystem;
	    USHORT  DllCharacteristics;
	    ULONG   SizeOfStackReserve;
	    ULONG   SizeOfStackCommit;
	    ULONG   SizeOfHeapReserve;
	    ULONG   SizeOfHeapCommit;
	    ULONG   LoaderFlags;
	    ULONG   NumberOfRvaAndSizes;
	    IMAGE_DATA_DIRECTORY DataDirectory[IMAGE_NUMBEROF_DIRECTORY_ENTRIES];
	} IMAGE_OPTIONAL_HEADER, *PIMAGE_OPTIONAL_HEADER;
	
	#define IMAGE_SIZEOF_SHORT_NAME              8
	
	#define IMAGE_SCN_MEM_EXECUTE 0x20000000
	#define IMAGE_SCN_MEM_READ 0x40000000
	#define IMAGE_SCN_MEM_WRITE 0x80000000
	
	typedef struct _IMAGE_SECTION_HEADER {
	    UCHAR   Name[IMAGE_SIZEOF_SHORT_NAME];
	    union {
	            ULONG   PhysicalAddress;
	            ULONG   VirtualSize;
	    } Misc;
	    ULONG   VirtualAddress;
	    ULONG   SizeOfRawData;
	    ULONG   PointerToRawData;
	    ULONG   PointerToRelocations;
	    ULONG   PointerToLinenumbers;
	    USHORT  NumberOfRelocations;
	    USHORT  NumberOfLinenumbers;
	    ULONG   Characteristics;
	} IMAGE_SECTION_HEADER, *PIMAGE_SECTION_HEADER;
	
	
	PIMAGE_DOS_HEADER dosheader = (PIMAGE_DOS_HEADER)binary;
	PIMAGE_FILE_HEADER fileheader = (PIMAGE_FILE_HEADER)((uint8_t*)dosheader + dosheader->e_lfanew + 4);
	if (memcmp((uint8_t*)fileheader-4, "PE\0\0", 4)) return;
	PIMAGE_OPTIONAL_HEADER optheader = (PIMAGE_OPTIONAL_HEADER)((uint8_t*)fileheader + sizeof(IMAGE_FILE_HEADER));
	
	uint8_t bits;
	if(0);
	else if (optheader->Magic == 0x010B) bits=32;
	else if (optheader->Magic == 0x020B) bits=64;
	else return;
	
	PIMAGE_SECTION_HEADER sections = (PIMAGE_SECTION_HEADER)((uint8_t*)optheader + sizeof(IMAGE_OPTIONAL_HEADER));
	for (uint16_t i=0;i<fileheader->NumberOfSections;i++)
	{
		if (sections[i].Characteristics & IMAGE_SCN_MEM_EXECUTE)
		{
			disassemble_section(binary + sections[i].PointerToRawData, sections[i].SizeOfRawData,
			                    optheader->ImageBase + sections[i].VirtualAddress, bits);
		}
	}
}



void disassemble_linux(uint8_t* binary, size_t len)
{
	typedef uint32_t Elf32_Addr;
	typedef uint16_t Elf32_Half;
	typedef uint32_t Elf32_Off;
	typedef  int32_t Elf32_Sword;
	typedef uint32_t Elf32_Word;
	
	typedef uint16_t Elf32_Section;
	
	#define EI_NIDENT (16)
	
	typedef struct
	{
	  unsigned char e_ident[EI_NIDENT];     /* Magic number and other info */
	  Elf32_Half    e_type;                 /* Object file type */
	  Elf32_Half    e_machine;              /* Architecture */
	  Elf32_Word    e_version;              /* Object file version */
	  Elf32_Addr    e_entry;                /* Entry point virtual address */
	  Elf32_Off     e_phoff;                /* Program header table file offset */
	  Elf32_Off     e_shoff;                /* Section header table file offset */
	  Elf32_Word    e_flags;                /* Processor-specific flags */
	  Elf32_Half    e_ehsize;               /* ELF header size in bytes */
	  Elf32_Half    e_phentsize;            /* Program header table entry size */
	  Elf32_Half    e_phnum;                /* Program header table entry count */
	  Elf32_Half    e_shentsize;            /* Section header table entry size */
	  Elf32_Half    e_shnum;                /* Section header table entry count */
	  Elf32_Half    e_shstrndx;             /* Section header string table index */
	} Elf32_Ehdr;
	
	#define SHT_NULL          0             /* Section header table entry unused */
	#define SHT_PROGBITS      1             /* Program data */
	#define SHT_SYMTAB        2             /* Symbol table */
	#define SHT_STRTAB        3             /* String table */
	#define SHT_RELA          4             /* Relocation entries with addends */
	#define SHT_HASH          5             /* Symbol hash table */
	#define SHT_DYNAMIC       6             /* Dynamic linking information */
	#define SHT_NOTE          7             /* Notes */
	#define SHT_NOBITS        8             /* Program space with no data (bss) */
	#define SHT_REL           9             /* Relocation entries, no addends */
	#define SHT_SHLIB         10            /* Reserved */
	#define SHT_DYNSYM        11            /* Dynamic linker symbol table */
	#define SHT_INIT_ARRAY    14            /* Array of constructors */
	#define SHT_FINI_ARRAY    15            /* Array of destructors */
	#define SHT_PREINIT_ARRAY 16            /* Array of pre-constructors */
	#define SHT_GROUP         17            /* Section group */
	#define SHT_SYMTAB_SHNDX  18            /* Extended section indeces */
	#define	SHT_NUM           19            /* Number of defined types.  */
	#define SHT_LOOS          0x60000000    /* Start OS-specific.  */
	#define SHT_GNU_ATTRIBUTES 0x6ffffff5   /* Object attributes.  */
	#define SHT_GNU_HASH      0x6ffffff6    /* GNU-style hash table.  */
	#define SHT_GNU_LIBLIST   0x6ffffff7    /* Prelink library list */
	#define SHT_CHECKSUM      0x6ffffff8    /* Checksum for DSO content.  */
	#define SHT_LOSUNW        0x6ffffffa    /* Sun-specific low bound.  */
	#define SHT_SUNW_move     0x6ffffffa
	#define SHT_SUNW_COMDAT   0x6ffffffb
	#define SHT_SUNW_syminfo  0x6ffffffc
	#define SHT_GNU_verdef    0x6ffffffd    /* Version definition section.  */
	#define SHT_GNU_verneed   0x6ffffffe    /* Version needs section.  */
	#define SHT_GNU_versym    0x6fffffff    /* Version symbol table.  */
	#define SHT_HISUNW        0x6fffffff    /* Sun-specific high bound.  */
	#define SHT_HIOS          0x6fffffff    /* End OS-specific type */
	#define SHT_LOPROC        0x70000000    /* Start of processor-specific */
	#define SHT_HIPROC        0x7fffffff    /* End of processor-specific */
	#define SHT_LOUSER        0x80000000    /* Start of application-specific */
	#define SHT_HIUSER        0x8fffffff    /* End of application-specific */
	
	#define SHF_EXECINSTR        (1 << 2)   /* Executable */
	typedef struct
	{
	  Elf32_Word    sh_name;                /* Section name (string tbl index) */
	  Elf32_Word    sh_type;                /* Section type */
	  Elf32_Word    sh_flags;               /* Section flags */
	  Elf32_Addr    sh_addr;                /* Section virtual addr at execution */
	  Elf32_Off     sh_offset;              /* Section file offset */
	  Elf32_Word    sh_size;                /* Section size in bytes */
	  Elf32_Word    sh_link;                /* Link to another section */
	  Elf32_Word    sh_info;                /* Additional section information */
	  Elf32_Word    sh_addralign;           /* Section alignment */
	  Elf32_Word    sh_entsize;             /* Entry size if section holds table */
	} Elf32_Shdr;
	
	typedef struct
	{
	  Elf32_Word    st_name;                /* Symbol name (string tbl index) */
	  Elf32_Addr    st_value;               /* Symbol value */
	  Elf32_Word    st_size;                /* Symbol size */
	  unsigned char st_info;                /* Symbol type and binding */
	  unsigned char st_other;               /* Symbol visibility */
	  Elf32_Section st_shndx;               /* Section index */
	} Elf32_Sym;
	
	Elf32_Ehdr* elfheader = (Elf32_Ehdr*)binary;
	//Elf32_Phdr* progheaders = (Elf32_Phdr*)(binary + elfheader->e_phoff);
	//for (Elf32_Half phnum=0;phnum<elfheader->e_phnum;phnum++)
	//{
	//	printf("P %.8X %.8X\n", progheaders[phnum].p_offset, progheaders[phnum].p_vaddr);
	//}
	Elf32_Shdr* secheaders = (Elf32_Shdr*)(binary + elfheader->e_shoff);
	
//	Elf32_Sym* dynsym = NULL;
//Elf32_Sym* dynsym_i;
//uint32_t dynsym_p;
//
//	char* dynstr = NULL;
	for (Elf32_Half shnum=0;shnum<elfheader->e_shnum;shnum++)
	{
		uint8_t* secstart = binary + secheaders->sh_offset;
//		//tiny: printf() imported to 0x00000000_00601028, section .got.plt type SHT_PROGBITS
//		//zsnes.elf: execve() imported to 0x08322254
//		if (secheaders->sh_type == SHT_DYNSYM)
//		{
//			dynsym = (Elf32_Sym*)secstart;
//			//printf("SY: %.8X\n",secheaders->sh_addr);
//dynsym_i = dynsym;
//dynsym_p = secheaders->sh_addr;
//		}
//		if (secheaders->sh_type == SHT_STRTAB && !dynstr) dynstr = (char*)secstart;
//printf("S %.8X %.8X\n",secheaders->sh_type,secheaders->sh_addr);
//		
//		//printf("%.8X\n",secheaders->sh_type);
		
		if (secheaders->sh_type == SHT_PROGBITS && secheaders->sh_flags&SHF_EXECINSTR)
		{
			disassemble_section(secstart, secheaders->sh_size, secheaders->sh_addr, 32);
		}
		secheaders++;
	}

//for (int i=0;true;i++)
//{
//printf("'%s' %.8X %.4X %.8X\n", dynstr + dynsym->st_name, dynsym->st_value, dynsym->st_shndx, ((char*)dynsym-(char*)dynsym_i)+dynsym_p);
//dynsym++;
//}
}
 //0 .interp            00000013  08048154  08048154  00000154  2**0  CONTENTS, ALLOC, LOAD, READONLY, DATA
 //1 .note.ABI-tag      00000020  08048168  08048168  00000168  2**2  CONTENTS, ALLOC, LOAD, READONLY, DATA
 //2 .note.gnu.build-id 00000024  08048188  08048188  00000188  2**2  CONTENTS, ALLOC, LOAD, READONLY, DATA
 //3 .gnu.hash          00000020  080481ac  080481ac  000001ac  2**2  CONTENTS, ALLOC, LOAD, READONLY, DATA
 //4 .dynsym            00000050  080481cc  080481cc  000001cc  2**2  CONTENTS, ALLOC, LOAD, READONLY, DATA
 //5 .dynstr            0000004c  0804821c  0804821c  0000021c  2**0  CONTENTS, ALLOC, LOAD, READONLY, DATA
 //6 .gnu.version       0000000a  08048268  08048268  00000268  2**1  CONTENTS, ALLOC, LOAD, READONLY, DATA
 //7 .gnu.version_r     00000020  08048274  08048274  00000274  2**2  CONTENTS, ALLOC, LOAD, READONLY, DATA
 //8 .rel.dyn           00000008  08048294  08048294  00000294  2**2  CONTENTS, ALLOC, LOAD, READONLY, DATA
 //9 .rel.plt           00000018  0804829c  0804829c  0000029c  2**2  CONTENTS, ALLOC, LOAD, READONLY, DATA
//10 .init              00000023  080482b4  080482b4  000002b4  2**2  CONTENTS, ALLOC, LOAD, READONLY, CODE
//11 .plt               00000040  080482e0  080482e0  000002e0  2**4  CONTENTS, ALLOC, LOAD, READONLY, CODE
//12 .text              00000192  08048320  08048320  00000320  2**4  CONTENTS, ALLOC, LOAD, READONLY, CODE
//13 .fini              00000014  080484b4  080484b4  000004b4  2**2  CONTENTS, ALLOC, LOAD, READONLY, CODE
//14 .rodata            0000000d  080484c8  080484c8  000004c8  2**2  CONTENTS, ALLOC, LOAD, READONLY, DATA
//15 .eh_frame_hdr      0000002c  080484d8  080484d8  000004d8  2**2  CONTENTS, ALLOC, LOAD, READONLY, DATA
//16 .eh_frame          000000b0  08048504  08048504  00000504  2**2  CONTENTS, ALLOC, LOAD, READONLY, DATA
//17 .init_array        00000004  08049f08  08049f08  00000f08  2**2  CONTENTS, ALLOC, LOAD, DATA
//18 .fini_array        00000004  08049f0c  08049f0c  00000f0c  2**2  CONTENTS, ALLOC, LOAD, DATA
//19 .jcr               00000004  08049f10  08049f10  00000f10  2**2  CONTENTS, ALLOC, LOAD, DATA
//20 .dynamic           000000e8  08049f14  08049f14  00000f14  2**2  CONTENTS, ALLOC, LOAD, DATA
//21 .got               00000004  08049ffc  08049ffc  00000ffc  2**2  CONTENTS, ALLOC, LOAD, DATA
//22 .got.plt           00000018  0804a000  0804a000  00001000  2**2  CONTENTS, ALLOC, LOAD, DATA
//23 .data              00000008  0804a018  0804a018  00001018  2**2  CONTENTS, ALLOC, LOAD, DATA
//24 .bss               00000004  0804a020  0804a020  00001020  2**0  ALLOC
//25 .comment           00000024  00000000  00000000  00001020  2**0  CONTENTS, READONLY

































int main(int argc, char * argv[])
{
	if (!argv[1])
	{
		puts("usage: ./mass-dis file\nexample: ./mass-dis xkas.exe");
		return 0;
	}
	
	FILE * f=fopen(argv[1], "rb");
	fseek(f, 0, SEEK_END);
	size_t len=ftell(f);
	fseek(f, 0, SEEK_SET);
	uint8_t* data=(uint8_t*)malloc(len);
	fread(data, 1,len, f);
	fclose(f);
	
	if (data[0]=='M' && data[1]=='Z') disassemble_windows(data, len);
	if (data[0]=='\x7F' && data[1]=='E' && data[2]=='L' && data[3]=='F') disassemble_linux(data, len);
	
	//disassemble_section();
	
	return 0;
}
